/*! @file
 * @brief
 * implementation of (invariant) CRelation methods
 * that can be called by servers and clients
 */

#include <string>
#include <iostream>
#include <cmath>
#include <gmp.h>
#include "utils.H"

#include "qsieve.H"

using namespace std;

extern int PrimeNumbers[StaticFactorbase::MaxSize];
extern mpz_t n;


ostream& operator<< (ostream& ostr, const CRelation& GL)
{
  if (GL.Relation_dense) ostr << "dense"; else ostr << "sparse";
  ostr << "1=";
  
  char *str;
  str = new char [mpz_sizeinbase(GL.Delta,10)+2]; mpz_get_str(str, 10, GL.Delta);
  ostr << str << "^2";
  delete [] str; 
  
  if (GL.Relation_dense)
    { 
      for (int i=GL.Relation_dense->first(); i>=0; i=GL.Relation_dense->next(i))
	ostr << "*p" << i;
    }
  else
    {
      const CTinyFBsizetypeVector& R = *GL.Relation_sparse;
      for (int pos=0; pos<R.size(); ++pos)
	ostr << "*p" << R[pos];
    }
  
  return ostr;
}


void CRelation::convert_Relation_to_dense()
{
  //cout << "Converting relation from sparse to dense representation." << endl;
  if (Relation_dense)
    { cerr << "Relation is already dense!" << endl; return; }
  Relation_dense = new myBitString;
  for (int pos=Relation_sparse->size()-1; pos>=0; --pos)
    Relation_dense->set((*Relation_sparse)[pos]);
  //Relation_sparse->erase(Relation_sparse->begin(),Relation_sparse->end());
  delete Relation_sparse; Relation_sparse=NULL; // delete sparse representation
}

void CRelation::convert_Relation_to_sparse()
{
  //cout << "Converting relation from dense to sparse representation." << endl;
  if (Relation_sparse)
    { cerr << "Relation is already sparse!" << endl; return; }
  Relation_sparse = new CTinyFBsizetypeVector(Relation_dense->count(1));
  for (int i=Relation_dense->first(); i>=0; i=Relation_dense->next(i))
    Relation_sparse->fast_append(i);
  //Relation_dense->clear();
  delete Relation_dense; Relation_dense=NULL; // delete dense representation
}


void CRelation::combine(const CRelation& GL2)
  // GL2 as reference, because this more efficient
  // compute symmetric difference
{
  //cout << "combining relations..." << endl;
  //cout << "Gl.1: "; SanityCheck();
  //cout << "Gl.2: "; GL2.SanityCheck();
  
  /*
    Compute symmetric difference of the two relation sets.
    Multiply common part (intersection) into square root Delta.
  */
  mpz_set_ui(x,1); // zur Beschleunigung Faktoren in x aufsammeln, bis sie grer als n werden

  if ( GL2.Relation_dense ) // GL2.Relation is dense
    {
      if (!Relation_dense) // if relation isn't dense already,
	convert_Relation_to_dense(); // make it dense...
      
      for (int i=GL2.Relation_dense->last(); i>=0; i=GL2.Relation_dense->prev(i))
	if (Relation_dense->test(i))
	  if (i>0)
	    {
	      mpz_mul_ui(x,x,PrimeNumbers[i]);
	      if (mpz_cmp(x,n)>=0)
		{
		  mpz_mod(x,x,n); mpz_mul(Delta,Delta,x);
		  mpz_mod(Delta,Delta,n); mpz_set_ui(x,1);
		}
	    } //else mpz_neg(Delta,Delta);
      
      (*Relation_dense)._xor(*GL2.Relation_dense); // symmetric difference
    }
  else /* GL2.Relation is sparse */
    if (Relation_dense) /* and Relation is dense */
      { 
        // compute symmetric difference in dense representation
	// (whereas the other vector remains sparse)
	for (int pos=0; pos<GL2.Relation_sparse->size(); ++pos)
	  {
	    if (Relation_dense->test((*GL2.Relation_sparse)[pos]))
	      if ((*GL2.Relation_sparse)[pos]>0)
		{
		  mpz_mul_ui(x,x,PrimeNumbers[(*GL2.Relation_sparse)[pos]]);
		  if (mpz_cmp(x,n)>=0)
		    {
		      mpz_mod(x,x,n); mpz_mul(Delta,Delta,x);
		      mpz_mod(Delta,Delta,n); mpz_set_ui(x,1);
		    }
		} //else mpz_neg(Delta,Delta);
	    Relation_dense->invert((*GL2.Relation_sparse)[pos]);
	  }
      }
    else
      {
        // compute symmetric difference in sparse representation
	int p1 = 0;
	int p2 = 0;
	static CTinyFBsizetypeVector neueMenge(StaticFactorbase::Size());
	neueMenge.reset();
        const CTinyFBsizetypeVector& R = *Relation_sparse;
	const CTinyFBsizetypeVector& R2 = *GL2.Relation_sparse;
	while (p2<R2.size() && p1<R.size())
	  {
	    while (R2[p2]<R[p1])
	      {
		neueMenge.fast_append(R2[p2]); ++p2;
		if (p2==R2.size()) goto done;
	      }
	    while (R[p1]<R2[p2])
	      {
		neueMenge.fast_append(R[p1]); ++p1;
		if (p1==R.size()) goto done;
	      }
	    while (R2[p2]==R[p1])
	      {
		if (R[p1]>0)
		  {
		    mpz_mul_ui(x,x,PrimeNumbers[R[p1]]);
		    if (mpz_cmp(x,n)>=0)
		      {
			mpz_mod(x,x,n); mpz_mul(Delta,Delta,x);
			mpz_mod(Delta,Delta,n); mpz_set_ui(x,1);
		      }
		  } //else mpz_neg(Delta,Delta);
		++p1; ++p2;
		if (p2==R2.size()) goto done;
		if (p1==R.size()) goto done;      
	      }
	  }
      done:
	while (p2!=R2.size()) { neueMenge.fast_append(R2[p2]); ++p2; }
	while (p1!=R.size()) { neueMenge.fast_append(R[p1]); ++p1; }
	Relation_sparse->copy_from(neueMenge); // take over new set
        //optisize(); // optize memory usage (sparse<->dense, etc.)
      }
  
  // x contains (possibly) a remaining value, which has to be multiplied into Delta:
  mpz_mul(Delta,Delta,x); mpz_mod(Delta,Delta,n);
  
  // don't forget: the new delta has to contain the delta of the other
  // relation as well!
  mpz_mul(Delta,Delta,GL2.Delta); mpz_mod(Delta,Delta,n);
  
  relevant_factor=largest_factor_in_Relation();
  //cout << "combine finished." << endl;
}


// The following set of methods greatly improve multiple combination of relations.
//  idea: the update/combine of the exponents of the factors can
//   deferred up to the time, where the result is needed.
//   (In most situations there is really no need to this simultaneously.)

CRelation::TExponentArrayElement CRelation::ExponentArray[StaticFactorbase::MaxSize+32] = {0};
 // +32 because we want to allow fast routines to avoid costly range checks...

unsigned int CRelation::multi_combine_Counter = 0;
 
void CRelation::multi_combine_init()
{
  // initialize vector of exponents.
  // hint: If you can assure, that the last call of multi_combine_...
  //       was in fact a multi_combine_exit(), then you
  //       may skip calling multi_combine_init().
  if (!multi_combine_Counter) return;
  multi_combine_Counter=0;
  for (int i=0; i<StaticFactorbase::Size(); i++) ExponentArray[i]=0;
}

void CRelation::multi_combine_exit()
{
  if (!multi_combine_Counter) return; // nothing to do...
#ifdef VERBOSE
  cout << "Multicombine_exit after " << multi_combine_Counter
       << " calls." << endl;
#endif

  if (multi_combine_Counter>std::numeric_limits<TExponentArrayElement>::max())
   {
     MARK;
     cerr << "Elements of CRelation::ExponentArray may have exceeded limits," << endl
          << "you have to increase the type of CRelation::TExponentArrayElement" << endl
          << "and recompile the program to assure correct computation." << endl
          << "(Data computed so far are still correct and can be reused afterwards.)" << endl;
     exit(8);
   }

  optisize(); // optimize Size of vector (sparse<->dense, etc.)

  // multiply factors weighted by theis exponents into Delta:
  mpz_set_ui(y,1); // collect product in y
  ExponentArray[0]=0; // leave "prime number" "-1" out of consideration

  /* 
     implemented idea:
     sort vector of exponents descending by exponents,
     evaluate the product analogously to the Horner scheme.
     -> sorting method: bucketsort. ("Flag[i]" defines, whether an exponent is assoziated to i.)
     All bases belonging to an exponent e will be multiplied together in "Basis[e]".
  */

  bool* const Flag = new bool[StaticFactorbase::Size()];
  for (int i=0; i<StaticFactorbase::Size(); ++i) Flag[i]=false; // initialize
  mpz_t* const Basis = new mpz_t[StaticFactorbase::Size()];

  for (int i=1; i<StaticFactorbase::Size(); ++i)
   if (ExponentArray[i])
    {
      const int e=ExponentArray[i];
      ExponentArray[i]=0; // reset exponent

      if (e>=StaticFactorbase::Size())
       {
         cout << "Unexpected exponent " << e << "! Use normal exponentiation..." << endl;
         mpz_set_ui(y,PrimeNumbers[i]); mpz_powm_ui(y,y,e,n);
         mpz_mul(Delta,Delta,y);
         continue;
       }

      if (Flag[e])
       {
         // expand Basis
         mpz_mul_ui(Basis[e],Basis[e],PrimeNumbers[i]);
         if (mpz_cmp(Basis[e],n)>=0)
          { 
            //cout << "Basis erweitern, mod sinnvoll..." << endl;
            mpz_mod(Basis[e],Basis[e],n);
          }
       }
      else 
       { 
         // initialize Basis
         Flag[e]=true;
         mpz_init(Basis[e]); mpz_set_ui(Basis[e],PrimeNumbers[i]);
       }
    }

 /*
  Evaluate product analogously to Horner scheme,
  example: a^7 * b^4 * c^3 * d 
           = a^(7-4) * (a*b)^(4-3) * (a*b*c)^(3-1) * (a*b*c*d)
 */     

  unsigned int e2 = StaticFactorbase::Size()-1;
  Flag[0]=true; // be cautious for next loop...
  while (!Flag[e2]) --e2; // get highest set exponent

  mpz_set_ui(y,1);
  if (e2>1)
   for (unsigned int e=e2-1; e>0; --e)
    if (Flag[e])
     {
       mpz_mul(Basis[e],Basis[e],Basis[e2]);
       if (mpz_cmp(Basis[e],n)>=0)
        {
          //cout << "Auswertung A, mod sinnvoll" << endl; 
          mpz_mod(Basis[e],Basis[e],n);
        }
       if (e2-e>1) mpz_powm_ui(Basis[e2],Basis[e2],e2-e,n);
       mpz_mul(y,y,Basis[e2]);
       if (mpz_cmp(y,n)>=0)
        {
          //cout << "Auswertung B, mod sinnvoll" << endl; 
          mpz_mod(y,y,n);
        }
       mpz_clear(Basis[e2]); // do not forget: free this mpz-number
       e2=e;
     }

  if (e2>1) mpz_powm_ui(Basis[e2],Basis[e2],e2,n);
  if (e2>0)
   {
      mpz_mul(y,y,Basis[e2]); mpz_mod(y,y,n);
      mpz_clear(Basis[e2]); // do not forget: free this mpz number
   }

  delete [] Basis; delete [] Flag;
  mpz_mul(Delta,Delta,y); mpz_mod(Delta,Delta,n);
  multi_combine_Counter=0;
#ifdef VERBOSE
  cout << "Multi_combine_exit finished." << endl;
#endif
}

void CRelation::multi_combine_main(const CRelation& GL2)
  // bergabe von GL2 ber Referenz, um effizienter zu sein...
  /* symmetrische Differenzen ermitteln */
{
  //cout << "multi-combine relations..." << endl;
  ++multi_combine_Counter; // count number of calls

  /*
    Compute symmetric difference of the two relation sets.
    Memorize common part (intersection) by adding it to Exponentenfeld.
  */

  if ( GL2.Relation_dense ) /* GL2.Relation is dense */
    {
      if (!Relation_dense) // Relation ist nicht dense
	convert_Relation_to_dense(); // Relation in dichte Darstellung bringen

      /* Zunchsten werden die Bits der beiden Relationen
         (temporr!) komponentenweise durch "and" verknpft.
         Dadurch werden die Primzahlen ermittelt, die verknpft Quadrate
         ergeben. Diese knnen dann im Exponentenfeld bercksichtigt werden.
         In gewisser Weise wird die Bitstrings komponentenweise wie
         mit einem "Halbaddierer" verknpft.
          Additionsergebnis (durch xor) berschreibt anschlieend den
          Bitstring, whrend der bertrag im Expoentenfeld addiert wird.
       */
      Relation_dense->test_and_add_carry(*GL2.Relation_dense,ExponentArray);
      (*Relation_dense)._xor(*GL2.Relation_dense); // symmetrische Differenz bilden
    }
  else // GL2.Relation is sparse
    if (Relation_dense) // and Relation is dense
      { // compute symmetric difference in dense represention,
        // whereas the other vector is sparse
	const CTinyFBsizetypeVector& R2 = *GL2.Relation_sparse;
	for (int pos=0; pos<R2.size(); ++pos)
	  {
	    if (Relation_dense->test_and_invert(R2[pos])) ++ExponentArray[R2[pos]];
	  }
      }
    else
      { // compute symmetric difference in sparse representation
	int p1 = 0;
	int p2 = 0;
	static CTinyFBsizetypeVector neueMenge(StaticFactorbase::Size());
	neueMenge.reset();
	const CTinyFBsizetypeVector& R  = *Relation_sparse;
	const CTinyFBsizetypeVector& R2 = *GL2.Relation_sparse;
	while (p2<R2.size() && p1<R.size())
	  {
	    while (R2[p2]<R[p1])
	      {
		neueMenge.fast_append(R2[p2]); ++p2;
		if (p2==R2.size()) goto done;
	      }
	    while (R[p1]<R2[p2])
	      {
		neueMenge.fast_append(R[p1]); ++p1;
		if (p1==R.size()) goto done;
	      }
	    while (R2[p2]==R[p1])
	      {
		++ExponentArray[R[p1]];
		++p1; ++p2;
		if (p2==R2.size()) goto done;
		if (p1==R.size()) goto done;      
	      }
	  }
      done:
	while (p2!=R2.size()) { neueMenge.fast_append(R2[p2]); ++p2; }
	while (p1!=R.size()) { neueMenge.fast_append(R[p1]); ++p1; }
	Relation_sparse->copy_from(neueMenge); // take over new set
      }

  // don't forget: the new delta has to contain the delta of the other
  // relation as well!
  mpz_mul(Delta,Delta,GL2.Delta); mpz_mod(Delta,Delta,n);
  
  //-------------------------------------------------------------------
  // IMPORTANT: After the last call of multi_combine_main it is necessary
  // to call a closing multi_combine_exit!
  // Otherwise the values in "Exponentenfeld" would be left out!
  //-------------------------------------------------------------------

  relevant_factor=largest_factor_in_Relation();
  //cout << "finished combine." << endl;
}



CmpqsFactor CRelation::combine(istream &in, const streampos pos)
{
  const streampos merkfpos=in.tellg(); // memorize fileposition
  in.seekg(pos);
  CmpqsFactor f=combine(in);
  in.seekg(merkfpos); // restore memorized fileposition
  return f;
}

CmpqsFactor CRelation::multi_combine_main(istream &in, const streampos pos)
{
  const streampos merkfpos=in.tellg(); // memorize fileposition
  in.seekg(pos);
  CmpqsFactor f=multi_combine_main(in);
  in.seekg(merkfpos); // restore memorized fileposition
  return f;
}


int GetValue(istream &in)
{
  static int ahead_value = 0;
  if (ahead_value>0) { int w=ahead_value; ahead_value=0; return w; }

  char c;
  in >> c;
  if ( c>='A' && c<='Z' ) return c-('A'-1); // A=1,...,Z=26
  if ( c>='g' && c<='z' )
   {
     int h = (c-'g');
     ahead_value=(h%5)+1; // next "ahead read" value
     return (h/5)+1; 
   }
  if (c=='0') return 0; // as we have no leading zeros, '0' means 0.
  if (c=='$') return -1;
  if (c=='%') return -2;

  if (c==' ')
   {
     MARK;
     cerr << "GetValue: Read a space? How to handle this?" << endl;
     exit(1);
   }

  // otherwise the memorized number:
  int w;
  in.unget();
  in >> w;
  return w;
}

void PutValue(ostream &out, const int w)
{
  static int prev_value = 0;
  if (w<=5 && w>=1 && prev_value>0) // then output coded value
    {
      out << char( (prev_value-1)*5 + w-1 + 'g' );
      prev_value=0;
      return;
    }

  if (prev_value>0)
   {
     out << char(prev_value+('A'-1)); // output value
     prev_value=0;
   }

  if (w>26) 
   {
     out << w << " "; // output value without encoding
     return;
   }
  if (w>0)
   {
     if (w<=4) { prev_value=w; return; } // wait till next time...
     out << char(w+('A'-1)); // output value coded as char
     return;
   }
  switch (w)
   {
     case -2: out << "%"; break;
     case -1: out << "$"; break;
     case  0: out << "0"; break;
     default:
      MARK;
      cerr << "PutValue: don't know how to handle this..." << endl;
      cerr << "w=" << w << endl;
      exit(1);
   }
}

void CRelation::SanityCheck() const
{
  mpz_set_ui(y,1); 
  if (Relation_sparse)
    {
      for (int i=0; i<Relation_sparse->size(); ++i)
	{
	  if ((*Relation_sparse)[i]>0) 
	    mpz_mul_ui(y,y,PrimeNumbers[(*Relation_sparse)[i]]);
	  else mpz_neg(y,y);
	  mpz_mod(y,y,n);
	}
    }
  else
    {
      for (int i=Relation_dense->first(); i>=0; i=Relation_dense->next(i))
	{
	  if (i>0) mpz_mul_ui(y,y,PrimeNumbers[i]); else mpz_neg(y,y);
	  mpz_mod(y,y,n);
	}
    }
  
  mpz_invert(x,Delta,n); mpz_powm_ui(x,x,2,n);
  
  cout << "checking for congruency: ";
  cout << x << " = " << y << " (mod N)" << endl;
  if (mpz_cmp(x,y)!=0) 
    {
      cout << "Error! Values are NOT congruent!" << endl;
      exit(1);
    }
  
}
