/*! @file
 * @brief
 * mpqs sieving stuff, header file
 */

#ifndef SIEVING_HEADER_
#define SIEVING_HEADER_

#include <qsieve-fwd.H>
#include <cmath>


/*!
   @short
   size of sieve interval in cacheable memory for fast access

   The logical sieve interval is sieved in separate smaller parts, because
   memory access in cache can be up to 200 times faster than uncached memory
   access (cost of a cachemiss). The value of @p PhysicalSieveSize is
   ideally the amount of memory that fits into processor cache while not
   replacing too much other cacheable data.

   The relations are sieved in logical intervals [-M,M] using a loop of
   physical intervals of PhysicalSieveSize.
   PhysicalSieveSize must be set at compile time and is constant.

   If cache effects wouldn't matter, then a greater value of
   PhysicalSieveSize would be better than a smaller one. But cache effects
   do matter! A cache miss can cost between 10 and 200 instruction cycles!

   Smaller sieve sizes increase computational overhead, but they
   also increase cache efficency.

   However, you should use a multiple of 1024 to be on the safe side.
   If optimized assembler code comes into play, this contraint MUST be met!
 */
const int PhysicalSieveSize = 64*1024; // =65536


/*!
   Constraint for AMD-3DNow! specific code: abs(LogicalSieveSize)<=2^23 (because of float precision)

   In contrast to "PhysicalSieveSize", LogicalSieveSize is a value that is
   less dependent on the hardware. You should choose the (logical) SieveSize
   according to size of the number to factorize. The default value should be
   modified before a factorization is started.

   The relative rate of hits in the multiple polynomial quadratic sieve
   algorithm (MPQS) is higher for smaller intervals, but however, the cost
   of changing polynomials, initialization, and calculation of roots is
   higher, when polynomials are switched more often.
*/
extern int LogicalSieveSize; // sieving interval will be [-LogicalSieveSize,LogicalSieveSize] for each MPQS polynomial



// !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
// !!! IMPORTANT !!! #define and typedef MUST match !!!
// !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
#ifndef TSIEVEELEMENTSIZE
#warning "TSIEVEELEMENTSIZE not defined. Using default value: 1"
#define TSIEVEELEMENTSIZE 1 /* 1 = signed char, 2 = short signed int, 4 = signed int, 0 = no assembler inline code (=generic code) */
#endif

// if ASM386 is defined, then TSIEVEELEMENTSIZE defines the type which is used in inline-assembler subroutines

#if TSIEVEELEMENTSIZE == 0
 typedef signed char TSieveElement; // must be signed!
 #define log_SieveEntryMultiplier 1 /* sieve entries are casted to TSiebelement using this multiplier */
 #ifdef ASM386
  #warning "assembly code is disabled for sieving!"
 #endif
#elif TSIEVEELEMENTSIZE == 1
 typedef signed char TSieveElement; // must be signed!
 #define log_SieveEntryMultiplier 1 /* sieve entries are casted to TSiebelement using this multiplier */
#elif TSIEVEELEMENTSIZE == 2
 typedef signed short int TSieveElement; // must be signed!
 #define log_SieveEntryMultiplier 16 /* sieve entries are casted to TSiebelement using this multiplier */
#elif TSIEVEELEMENTSIZE == 4
 typedef signed int TSieveElement; // must be signed!
 #define log_SieveEntryMultiplier 256 /* sieve entries are casted to TSiebelement using this multiplier */
#else
 #error "invalid value for TSIEVEELEMENTSIZE"
#endif

/*
  The sieve needs less memory (and less cache), if TSieveElement is chosen
  of the smallest possible type. Sieving is faster for short types.

  On the other hand, the granularity to detect relations is worse for
  shorter datatypes than for longer ones.

  You have also to consider processor specific effects (eg. alignment penalties)!

  Following datatypes are possible:
    - signed char (1 byte)
    - short signed int (2 bytes)
    - signed int (4 bytes)
    - float, double (floating point types, several (>3) bytes)

  Size of types may vary between systems. I haven't checked, whether
  the program works for other contraints than
    int >= 4 bytes, short int >= 2 bytes.

  Important remark:
   If you use floating point types for TSieveElement (which I don't
   recommend anymore), then calling "ceil" is counterproductive (and such
   calls should be removed in the source code).  But there is really no
   reason, why one should do sieving using floating point types.
*/

/*
  about: log_SieveEntryMultiplier <VALUE>
  (-> refer TSIEVEELEMENTSIZE)

  Sieve entries are casted to TSieveElement using this multiplier to gain a
  better accuracy.  If the TSieveElement is of floating point type, then
  log_SieveEntryMultiplier can be set to 1. In all other cases you should
  set it to the highest possible value, such that the sieving threshold (for
  numbers usually to factorize) fits into TSieveElement.

  In order not to miss any detection of a relation when using short datatypes,
  you must use "ceil" while subtracting hits in the sieve. (Thereby the
  sieve gets close meshed.)
*/



#ifdef IS_SERVER

 // The server does no sieving,
 // but the server source code needs some information about how sieving
 // is done. Additionally the server sources contain some code that is
 // depending on data structures defined here. We could eliminate these
 // lines by conditional compiling (#ifdef IS_SERVER), but this would
 // be to bloated. Instead we provide basic dummy classes which produce
 // dead code, if compiled in SERVER-mode.

template <typename T> class CDummyArray
{
 private:
  T dummy;
 public:
  inline T& operator[] (const int)
   {
     return dummy;
   }
  inline const T& operator[] (const int) const
   {
     return dummy;
   }
};

class SieveControl
{
 private:
  static const int FBLowestStartIndex = 5;
  /* This is the first index of the factorbase to sieve with.
     This value needs to be greater than 1, because the first two factors of
     the factorbase are -1 and (probably) 2. Fore these two numbers as well
     as for multiplier of kN we are uable to sieve. (Well, at least it would
     be more difficult.) For a fast, but dirty sieving you can choose a
     greater value as well.  But the higher the value is, the more relations
     remain undetected...
   */

  inline static const double DirtyFactor(const int i)
   {
     if (i<FBLowestStartIndex+4) return 1.0;
     return 1.0/sqrt(i); // a function slowly going towards 0
   }
  /*
     DirtyFactor denotes the weight, with which the i-th prime number of the
     static factor base will hin into the sieve, if the hit is "dirty".

     Example:
      If you sieve the normal way using a prime number p, then you have a hit
      every p-th position in the sieve. This hit will be weighted by log(p).

      If  p is "dirty" then a hit will be weighted using DirtyFactor*log(p) for
      *every* sieve entry! Only if a relation is detected later, then this
      error (dirty-sieving effect) will be corrected to check, whether the
      hit would be a hiot under normal circustances. -- Doing this speeds up
      the sieving because less primes (especially the lower ones) need to be
      sieved at the cost of correcting the effect later. As normally only a
      small fraction of the sieve entries lead to relations, this should'nt
      be a problem if the dirty prime numbers are dynamically adaped to the
      rate of hits.

      A DirtyFactor of 1.0 behaves neutral compared with normal sieving except
      for (potential) rounding errors, but it spoeds up the sieving.

      The smaller the DirtyFactor, the more prime numbers can be dirty.
      (However, the rate of detection per sieve interval decreases in favour
      of sieving speed.)

      A DirtyFactor of 1/n behaves roughly as if each n-th dirty prime
      number would be a member in every relation.
   */

  static const int FBStartIndex(void) { return FBLowestStartIndex; }

  static CDummyArray<TSieveElement> log_PrimeNumbers;
  static CDummyArray<TSieveElement> log_PrimeNumbers_dirty;

 public:
  friend void compute_StaticFactorbase();
  friend void determine_best_kN();
};

CDummyArray<TSieveElement> SieveControl::log_PrimeNumbers;
CDummyArray<TSieveElement> SieveControl::log_PrimeNumbers_dirty;

CDummyArray<TSieveElement> log_Primzahl_of_PrimePowers;


#else /* IS_SERVER not defined */

// these are the "full" declarations
#include <iosfwd>
#include <gmp.h>
#include "utils.H"
#include "mpz_wrapper.H"
using namespace my_mpz_wrapper;

using std::ostream;
using std::istream;

using std::cout;
using std::cerr;


extern int SieveOffset;
/* In general the logical sieve interval for each mpqs polynomial will exceed
   the size of the physical sieve. The logical sieve has to be sieved in
   several parts/steps. SieveOffset is the offset of the logical sieve at
   the starting point of a physical sieve.
 */


extern TSieveElement SieveArray_[PhysicalSieveSize+64+8] __attribute__ ((aligned (64))); // Speicher fr das Sieb
TSieveElement* const SieveArray = &SieveArray_[64];
// SieveArray[-64..-1] are -1 (for fast value-based end-of-loop range detections)
// SieveArray[0..PhysicalSieveSize-1] contain actually sieve values
// SieveArray[PhysicalSieveSize..PhysicalSieveSize+8] contain dummy values


extern TSieveElement log_Primzahl_of_PrimePowers[StaticFactorbase::max_additional_Powers];
// converted logarithms for further powers of prime numbers


extern int PrimeNumbers[StaticFactorbase::MaxSize];
extern int Delta_of_PrimeNumbers[StaticFactorbase::MaxSize][2];
extern mpz_t n, kN, r;
extern double Factor_Threshold;
extern int biggest_Prime_in_Factorbase;
extern int NumberOf_more_PrimePowers;
extern int PrimePowers[StaticFactorbase::max_additional_Powers];
extern int Delta_of_PrimePowers[StaticFactorbase::max_additional_Powers][2];


inline TSieveElement log_DynamicLargePrime (const int p)
{
#if defined(ASM386) && TSIEVEELEMENTSIZE == 1 && log_SieveEntryMultiplier == 1
 #ifdef DEBUG
  #warning "assembler i386 TSIEVEELEMENT log lookup enabled"
 #endif
  // we assume 8103<p<=1318815734 (=e^9)
  register TSieveElement logval;
  asm ( \
       "movb $21,%[L] \n\t" \
       "cmpl $485165195,%[P] # exp(20) \n\t" \
       "sbbb $0,%[L] \n\t" \
       "cmpl $178482300,%[P] # exp(19) \n\t" \
       "sbbb $0,%[L] \n\t" \
       "cmpl $65659969,%[P] # exp(18) \n\t" \
       "sbbb $0,%[L] \n\t" \
       "cmpl $24154952,%[P] # exp(17) \n\t" \
       "sbbb $0,%[L] \n\t" \
       "cmpl $8886110,%[P] # exp(16) \n\t" \
       "sbbb $0,%[L] \n\t" \
       "cmpl $3269017,%[P] # exp(15) \n\t" \
       "sbbb $0,%[L] \n\t" \
       "cmpl $1202604,%[P] # exp(14) \n\t" \
       "sbbb $0,%[L] \n\t" \
       "cmpl $442413,%[P] # exp(13) \n\t" \
       "sbbb $0,%[L] \n\t" \
       "cmpl $162754,%[P] # exp(12) \n\t" \
       "sbbb $0,%[L] \n\t" \
       "cmpl $59874,%[P] # exp(11) \n\t" \
       "sbbb $0,%[L] \n\t" \
       "cmpl $22026,%[P] # exp(10) \n\t" \
       "sbbb $0,%[L]" \
       : [L] "=&q" (logval) : [P] "r" (p) : "cc");
#ifdef DEBUG
  // remark: do not use std::log(), as it might be ambiguous for some
  // versions of g++ (-> see utils.H)
  if (logval!=ceil(log_SieveEntryMultiplier*log(p)))
   {
     MARK;
     cout << "P: " << p << " log:" << static_cast<int>(logval) << "  " << ceil(log_SieveEntryMultiplier*log(p)) << endl;
     exit(1);
   }
#endif
  return logval;
#else
  // generic version
  return static_cast<TSieveElement>(ceil(log_SieveEntryMultiplier*log(p)));
#endif
}


/// provide adaptive control mechanism to improve sieving
class SieveControl
{
 private:
  static const int FBLowestStartIndex = 5;
  /* ab diesem Primzahlindex wird gesiebt.
     Der Wert mu grer als 1 sein, da die ersten beiden Faktoren
     der Primzahlbasis -1 und mglicherweise die 2 sind. Fr diese
     beiden Zahlen sowie dem Vorfaktor von kN kann nicht (jedenfalls nicht
     so einfach) gesiebt werden.
     Fr ein schnelles, aber unsauberes dirty-sieving kann auch ein
     grerer Wert gewhlt werden. Aber zu hohe Werte lassen dann auch
     viele Relationen durchs Sieb fallen...
   */

  inline static const double DirtyFactor(const int i)
   {
     if (i<FBLowestStartIndex+4) return 1.0;
     return 1.0/sqrt(i); // a function slowly going towards 0
   }
  /*
     DirtyFactor denotes the weight, with which the i-th prime number of the
     static factor base will hin into the sieve, if the hit is "dirty".

     Example:
      If you sieve the normal way using a prime number p, then you have a hit
      every p-th position in the sieve. This hit will be weighted by log(p).

      If  p is "dirty" then a hit will be weighted using DirtyFactor*log(p) for
      *every* sieve entry! Only if a relation is detected later, then this
      error (dirty-sieving effect) will be corrected to check, whether the
      hit would be a hiot under normal circustances. -- Doing this speeds up
      the sieving because less primes (especially the lower ones) need to be
      sieved at the cost of correcting the effect later. As normally only a
      small fraction of the sieve entries lead to relations, this should'nt
      be a problem if the dirty prime numbers are dynamically adaped to the
      rate of hits.

      A DirtyFactor of 1.0 behaves neutral compared with normal sieving except
      for (potential) rounding errors, but it spoeds up the sieving.

      The smaller the DirtyFactor, the more prime numbers can be dirty.
      (However, the rate of detection per sieve interval decreases in favour
      of sieving speed.)

      A DirtyFactor of 1/n behaves roughly as if each n-th dirty prime
      number would be a member in every relation.
   */

  static TSieveElement log_PrimeNumbers[StaticFactorbase::MaxSize];
  static TSieveElement log_PrimeNumbers_dirty[StaticFactorbase::MaxSize];
    // Hier stehen die konvertierten Logarithmen der Primzahlen aus der Faktorbasis

  static TSieveElement log_Threshold; // wird aber erst durch "Grundinitialisierung_des_Siebs" initialisiert
  static int myFBStartIndex;

  static signed int DirtynessPegel;

public:
  inline static const TSieveElement GetLogThreshold(void) { return log_Threshold; }
  inline static const int FBStartIndex(void) { return myFBStartIndex; }
  static void compute_SieveThreshold(void);
  static const bool Hit_after_DirtyCorrection(const signed int SievePos, TSieveElement Value);

  inline static const TSieveElement PrimzahlSieveWeight(const int i)
   {
     // Gibt die Wichtung der Primzahl im Sieb zurck; blicherweise ist
     // dies der (skalierte) Logarithmus, der in der Tabelle abgelegt wurde. 
     // Diese Funktion dient dazu, den Zugriff read-only zu kapseln.
     return log_PrimeNumbers[i];
   }
  inline static const TSieveElement PrimzahlSieveWeight_dirty(const int i)
   {
     // Gibt die Wichtung der Primzahl im Sieb zurck wenn diese als dirty
     // gekennzeichnet ist, d.h. auf ein Sieben mit dieser Zahl verzichtet
     // wird; dieser Wert wird dann fr alle Elemente des Siebs einheitlich
     // als Treffer gewichtet (statt tatschlich zu sieben); blicherweise
     // ist dies ein (skalierter) Logarithmus, der in der Tabelle abgelegt
     // wurde. Diese Funktion dient dazu, den Zugriff read-only zu kapseln.
     return log_PrimeNumbers_dirty[i];
   }

private:
  static void IncreaseDirtyness(void)
   {
     // Threshold absenken und Index fr erste zu siebende Zahl inkrementieren
     if (myFBStartIndex>(StaticFactorbase::Size()>>2)) return; // das geht leider nicht...
     log_Threshold-=PrimzahlSieveWeight_dirty(myFBStartIndex);
     ++myFBStartIndex;
#ifdef VERBOSE_INFO
     cout << "Dirtyness now " << myFBStartIndex << endl;
#endif
   }

  static void DecreaseDirtyness(void)
   {
     // Threshold hochsetzen und Index fr erste zu siebende Zahl dekrementieren
     if (myFBStartIndex<=FBLowestStartIndex) return; // das geht leider/wollen wir nicht...
     log_Threshold+=PrimzahlSieveWeight_dirty(--myFBStartIndex);
#ifdef VERBOSE_INFO
     cout << "Dirtyness now " << myFBStartIndex << endl;
#endif
   }

  static void RequestDirtyness(const signed int Level)
   {
     DirtynessPegel+=Level;
     if (DirtynessPegel>1000000)
      {
        DirtynessPegel=0;
        IncreaseDirtyness();
      } else
     if (DirtynessPegel<-1000000)
      {
        DirtynessPegel=0;
        DecreaseDirtyness();
      }
   }

 public:
  static inline void RequestMoreDirtyness(const signed int Level = 1000)
   {
     RequestDirtyness(Level);
   }
  static inline void RequestLessDirtyness(const signed int Level = 50)
   {
     RequestDirtyness(-Level);
   }

  friend void compute_StaticFactorbase();
  friend void determine_best_kN();
  friend void initialize_Sieve(void);
};


// helper class to encapsulate private data
class CSieveStaticFactors
{
 private:
  static int Local_Delta_of_PrimeNumbers[StaticFactorbase::MaxSize][2];
  static void init_LocalDeltas();
  static void do_sieving_StaticFactors();
  friend int main(const int argc, const char* const argv[]);
};


void do_scanning_Sieve(void);

#include <vector>
void GetStaticFactorHits(std::vector<unsigned int> &Factors, const signed int SievePos);


#endif /* IS_SERVER not defined */

#endif /* SIEVING_HEADER_ */
