/*! @file
 * @brief
 * implementation of class StaticRelations
 */

#include "StaticRelations.H"

CRelation* StaticRelations::GLS[StaticFactorbase::MaxSize] = {NULL};
int StaticRelations::Filling_GLS = 0;

StaticFactorbase::FBsizetype StaticFactorbase::Size_StaticFactorbase = 5000; // default size for factorbase
/*  
    Constraint for AMD-3DNow! specific code: Size_StaticFactorbase%2 == 0  (because of SIMD)
    Useful values are between 1000 and 50000 (depending on available memory and the number to factorize).
    Normally this value will be overwritten by a value given in the config file (see "tune_parameters").
*/


filebuf StaticRelations::FileBuffer;
ostream StaticRelations::StaticRelations_to_file(&StaticRelations::FileBuffer);
istream StaticRelations::StaticRelations_from_file(&StaticRelations::FileBuffer);


void StaticRelations::Load()
{
  // reading static relations from file
  //cout << "Reading static relations from file..." << endl;
  StaticRelations_from_file.seekg(0,ios::beg);
  while (StaticRelations_from_file.peek()=='G') 
   {
     CRelation* GL = new CRelation();
     GL->combine(StaticRelations_from_file);
     GL->optisize(); // Relation optimieren (dense<->sparse, etc.)
     GLS[GL->relevant_factor]=GL; ++Filling_GLS; // insert relation into linear system of equations
     StaticRelations_from_file.ignore(1,'\n');
   }
  StaticRelations_from_file.clear();
  //cout << "Load: " << Filling_GLS << " static relations read." << endl;
}

void StaticRelations::insert(CRelation *GL, const bool do_multi_combine_init /* =true */)
{
  /* Any new relation is inserted into the linear system of equations.
     If the relation is redundant or if it leads to a factorization,
     this will be detected automatically. */

#ifdef IS_CLIENT
  if (GL->relevant_factor >= 0) Filling_GLS++; // beim Client mangels Gleichungssystem: jede gefundene Relation
  delete GL;
  return;
#else  
  if (GL->relevant_factor==CRelation::dynamic_factor || GL->relevant_factor==CRelation::special_factor)
    { delete GL; return; } // save it to background storage and delete it in main memory
  
  if (do_multi_combine_init) GL->multi_combine_init();
  while (!GL->empty())
    {
      if (GL->relevant_factor>=StaticFactorbase::Size())
       {
         MARK;
         cerr << "value too big!" << endl;
         exit(1);
       }
      CRelation *pos = GLS[GL->relevant_factor];
      if (pos!=NULL)
	{
          // an entry for this relation already exists -> we can combine it!!
#if 1 /* toggle minimization of relations (1=on/0=off) */
          if (
	       (GL->Relation_sparse) && (pos->Relation_sparse) // both are sparse
               && GL->SizeOfRelation()<=pos->SizeOfRelation()
	       && GL->second_largest_factor_in_Relation()<=pos->second_largest_factor_in_Relation()
             )
	    {
              /* weniger Primzahlen in der Relation bedeuten schnelleres Abarbeiten.
	  	 Da beide Relationen quivalent sind, knnen sie problemlos so
		 getauscht werden, dass die platzsparende im Gleichungssystem verbleibt.
		 Andererseits bestimmt das Maximum der zweitgrten Faktoren zweier
		 Relationen die nchste Verknpfung. Also ist auch dies zu minimieren.
                 Beim Zielkonflikt wren beide Kriterien also abzuwgen. Dies geschieht
		 hier nicht. Nur wenn beide Kriterien zugleich die Lage nicht
                 verschlechtern, werden die Relationen getauscht.

                 Man beachte, dass der Tausch das Ergebnis der aktuellen Verknpfung
                 nicht verndert, sondern nur zuknftige Verknpfungen
	         beschleunigen kann bzw. den Speicherbedarf vermindert.
              */
	      GL->multi_combine_exit(); 
	      GLS[GL->relevant_factor]=GL;
	      std::swap(pos,GL); // Relationen tauschen
              // GL->multi_combine_init(); <- may be left out
	    }
#endif
          // biggest factor has to be eliminated
 	  GL->multi_combine_main(*pos); 
	}
      else
	{
	  GL->multi_combine_exit();
	  GLS[GL->relevant_factor]=GL; // insert relation into linear system of equations
	  Filling_GLS++; // one more relation in the system of linear equations
          GL->save(StaticRelations_to_file);
          StatusReport();
	  return;
	}
    }

  GL->multi_combine_exit();

  // the relation is redundant:
  // -> we cannot insert it into the static factorbase,
  // -> it can yield a factor instead!
  // -> compute quadratic congruency!
  const bool complete = GL->ComputeQuadraticCongruence();
  delete GL;
  if (complete)
   {
     // Game over, we have won!
     // leave the program...
     ExitManager::StopFactorization(); // try to stop the program successfully...
     exit(0); // this line should not be reached!
   }
#endif
}
