/*! @file
 * @brief
 * implementation of base class used in DLP-MPQS
 */

#include "Tfactor.H"
#include "modulo.H"
#include <cmath>
#include "qsieve-fwd.H"

using std::setw;
using std::setprecision;


// static class members
double CmpqsFactor::rejected_dlp_counter = 0.0;
mpz_t CmpqsFactor::DLP_Threshold; // Double-Large-Prime-Threshold; will be initialized in main()


// Pollard-rho to factorize Double-Large-Primes
const bool CmpqsFactor::DLP_get_using_pollard_rho(const mpz_t n)
{
  if (mpz_cmp(n,DLP_Threshold)>0)
   {
     //cout << "DLP_get: Threshold exceeded..." << endl;
     return false;
   }

  int runden=50000; // maximum #rounds for pollard-rho
  // hint: high value -> costly, low value: decreasing rate of detection

  mpz_t x,a,a2;
  mpz_init_set(x,n); mpz_init(a); mpz_init(a2);
  mpz_set_ui(a,1); mpz_set(a2,a);

  p1=0; p2=0; // failsafe defaults

  do
    {
      mpz_mul(a,a,a); mpz_add_ui(a,a,1); mpz_mod(a,a,n);
      mpz_mul(a2,a2,a2); mpz_add_ui(a2,a2,1); mpz_mod(a2,a2,n);
      mpz_mul(a2,a2,a2); mpz_add_ui(a2,a2,1); mpz_mod(a2,a2,n);
      mpz_sub(x,a2,a);
      mpz_gcd(x,x,n);
    } while (--runden && mpz_cmp_ui(x,1)==0);
  if (mpz_cmp_ui(x,1)!=0)
    {
      // check, whether both factors are prime and small enough
      if (mpz_cmp_ui(x,SingleLargePrime_Threshold)>=0) goto done;
      p1=mpz_get_ui(x);
      mpz_divexact(a,n,x);
      if (mpz_cmp_ui(a,SingleLargePrime_Threshold)>=0) goto done;
      p2=mpz_get_ui(a);
      // Hint: if we check here for primality, then we save the
      //       check later (when a factor is used for splitting).
      //       To be on the safe side, we do it now...
      if (!numtheory::probab_prime(p1)) { p1=0; goto done; }
      if (!numtheory::probab_prime(p2)) { p2=0; goto done; }
      if (p1>p2) std::swap(p1,p2);
    };
done:
  mpz_clear(a); mpz_clear(a2); mpz_clear(x);
  //if (runden==0) cout << "DLP_get: rounds exceeded." << endl;
  cout << "DLP_prho_get (" << 50000-runden << ") for " << n << ": " << p1 << "," << p2 << endl;
  
  //static unsigned int runden_ges = 0;
  //runden_ges+=50000-runden;
  //cout << "DLP_get: rounds needed: " << runden_ges << endl;

  return (p1!=0 && p2!=0);
}


// Shank's square forms to factorize Double-Large-Primes
const bool CmpqsFactor::DLP_get(const mpz_t n)
{
  const unsigned int runden = 50000; // maximum rounds, which will be tried
  // hint for tuning: more rounds cost more time, less rounds decrease rate of detection

  mpz_t mx,mr;
  mpz_init(mx); mpz_init(mr);

  p1=0; p2=0; // Vorbesetzung
  unsigned int SQUFOF_Vorfaktor = 1; // we assume this to be 1 or a very small prime number!

try_new_Vorfaktor:
  if (SQUFOF_Vorfaktor==1) mpz_sqrtrem(mx,mr,n);
  else
   {
     mpz_mul_ui(mx,n,SQUFOF_Vorfaktor);
     mpz_sqrtrem(mx,mr,mx);
   }
  const unsigned int sq = mpz_get_ui(mx);
  const unsigned int d = mpz_get_ui(mr);
  mpz_mul_ui(mx,mx,2); mpz_add_ui(mx,mx,1); mpz_sqrt(mx,mx);
  const unsigned int sq2sqN = mpz_get_ui(mx);

  unsigned int lindex=0;
  const unsigned int bound = 60;
  unsigned int list[bound];
  bool square_found=false;

  unsigned int Q,P2,Q2;
  unsigned int r=0, Q0=1, P1=sq, Q1=d;
  unsigned int runde1=0, runde2=0;

  if (d==0)
   {
     // perfect square!!
     p1=p2=sq; goto done;
   }

  for (runde1=1; runde1<runden; ++runde1)
   {
     //cout << "Runde" << runde1 << endl;
     Q=(sq+P1)/Q1;
     P2=sq-((sq+P1)%Q1); // equivalent to P2=Q*Q1-P1, but "/" and "%" probably use the same cpu instruction!
     if (P1>=P2) Q2=Q0+Q*(P1-P2); else Q2=Q0-Q*(P2-P1);

     // "unsigned int" has one more bit than "signed int", so overflows
     // should not occur at all,
     // but therefore now we have to take care for the sign!
     // Anyway: seldom overflows will cause no harm at all,
     // they only waste computing time (resulting in pollard rho doing
     // the job again...)
     // -> see below: "this line shouldn't be..."

     const unsigned int u = (Q1&1) ? Q1 : Q1>>1;
     Q0=Q1; Q1=Q2; P1=P2;
     if (u<sq2sqN && u>1)
      {
        list[lindex++]=u; 
        if (lindex>=bound)
         {
           cout << "SQUFOF: list exceeded (out of bound)..." << endl;
           break;
         }
      }
     if (runde1&1)
      {
        if ((Q1&3)>1)
          continue; // not a square
        if ((Q1&7)==5 || (Q1&9)==8)
          continue; // not a square (either)
        r=static_cast<unsigned int>(sqrt(Q1)); if (Q1!=r*r) continue; // not a square
        bool fl=false;
        for (unsigned int k=0; k<lindex; ++k)
         if (r==list[k]) { fl=true; break; }
        if (fl) continue; // Square not useful
        square_found=(r>1); break;
      }
   }

  if (!square_found)
   {
     //cout << "DLP-SQUFOF, Vorfaktor " << SQUFOF_Vorfaktor << ": No success." << endl;
     switch (SQUFOF_Vorfaktor)
      {
        case 1: SQUFOF_Vorfaktor=2; break;
        case 2: SQUFOF_Vorfaktor=3; break;
        default:
           cout << "DLP-SQUFOF: Fallback to pollard rho method." << endl;
           DLP_get_using_pollard_rho(n);
           goto done;
      }
     goto try_new_Vorfaktor;
   }

  //cout << "DLP-SQUFOF: square found in round " << runde1 << ": " << r << endl;

  double DQ,DQ0,DQ1,DQ2,DP1,DP2;
  mpz_set_ui(mr,P1); mpz_mul_ui(mr,mr,P1);

  if (SQUFOF_Vorfaktor==1) mpz_sub(mx,n,mr);
  else { mpz_mul_ui(mx,n,SQUFOF_Vorfaktor); mpz_sub(mx,mx,mr); }
  mpz_div_ui(mx,mx,r);
  DQ1=mpz_get_d(mx);
  DQ0=r; DP1=P1;

  for (runde2=1; runde2<runden; ++runde2)
   {
     DQ=floor((sq+DP1)/DQ1); DP2=DQ*DQ1-DP1;
     DQ2=DQ0+DQ*(DP1-DP2); DQ0=DQ1; DQ1=DQ2;
     if (DP1==DP2)
      {
        // factor found
        if ((DQ0/2)==floor(DQ0/2)) DQ0/=2;
        if ( SQUFOF_Vorfaktor>1 && (DQ0/SQUFOF_Vorfaktor)==floor(DQ0/SQUFOF_Vorfaktor) ) DQ0/=SQUFOF_Vorfaktor;

        if (DQ0==1.0)
         {
           //cout << "DLP-SQUFOF: Vorfaktor " << SQUFOF_Vorfaktor << " rediscovered; falling back to pollard rho..." << endl;
           DLP_get_using_pollard_rho(n); goto done;
         }

        // check, whether both factors are prime and small enough

        if (DQ0>=SingleLargePrime_Threshold) goto done;
        p1=static_cast<unsigned int>(DQ0);
        if (mpz_div_ui(mx,n,p1)) { cerr << "DLP_get-SQUFOF: weird factors! remainder??" << endl; exit(1); }
        if (mpz_cmp_ui(mx,SingleLargePrime_Threshold)>=0) goto done;
        p2=mpz_get_ui(mx);
        // Hint: if we check here for primality, then we save the
        //       check later (when a factor is used for splitting).
        //       To be on the safe side, we do it now...
        if (!numtheory::probab_prime(p1)) { p1=0; goto done; }
        if (!numtheory::probab_prime(p2)) { p2=0; goto done; }
        if (p1>p2) std::swap(p1,p2);
        //cout << "DLP-SQUFOF: factorization found!" << endl;
        goto done;
      }
     DP1=DP2;
   }
  MARK;
  cerr << "This line should'nt be executed!!! Overflow in unsigned int?? Negative Values??" << endl;
  DLP_get_using_pollard_rho(n); // fallback...

done:
  mpz_clear(mx); mpz_clear(mr);

#if defined(VERBOSE)
  cout << "DLP_get (SQUFOF," << SQUFOF_Vorfaktor << "): (" << runde1 << "," << runde2 << ") "
       << p1 << "," << p2 << endl;
#endif

#if 1 /* some statistics */
 static double good_dlp_counter = 0.0;
 static double bad_dlp_counter  = 0.0;
 if (p1&&p2) good_dlp_counter+=1.0; else bad_dlp_counter+=1.0;
 static time_t lastout = 0;
 if (time(NULL)>lastout+60)
  {
    lastout=time(NULL);
    cout << "DLP-SQUFOF: "
         << setw(6) << setprecision(5)
         << 100.0*rejected_dlp_counter/(rejected_dlp_counter+good_dlp_counter+bad_dlp_counter)
         << "% of DLP-candidates were rejected!" << endl;
    cout << "DLP-SQUFOF: "
         << setw(10) << setprecision(0) << good_dlp_counter 
         << " [" << setw(6) << setprecision(5)
         << 100.0*good_dlp_counter/(good_dlp_counter+bad_dlp_counter)
         << "%] of non-rejected DLP are good!" << endl;
  }
#endif

  return (p1&&p2);
}


istream& operator>> (istream &istr, CmpqsFactor &x)
{
   char s[50];
   istr >> setw(sizeof(s)) >> s;
   x.p1=0; x.p2=0;
   int i=0;
   while (s[i]!=0 && s[i]!='*') ++i;
   if (s[i]==0) 
    {
      x.p1=0; x.p2=atoi(s);
      //cout << "DLP: read a value" << endl;
    }
   else 
    if (s[i]=='*')
     {
       s[i]=0;
       x.p1=atoi(s);
       x.p2=atoi(&s[i+1]);
       //cout << "DLP: read two values" << endl;
     }
    else
     {
       cerr << "Reading DLP failed!" << endl;
     }
   //cout << "DLP:" << x.p1 << "*" << x.p2 << endl;
   return istr;
}
