/*! @file
 * @brief
 * contains some types and constants for qsieve.H and qsieve.cc
 */

#ifndef QSIEVE_FWD_HEADER_
#define QSIEVE_FWD_HEADER_

#include "utils.H"

const int server_port = 12364;
// server port for networked distributed factoring

const int XML_status_port = 12370;
// port for accessing XML status information


const int SingleLargePrime_Threshold = 1000000000;
// up to this this value the sieved single large primes are
// accepted as such. Only SLP (single large primes) up to this
// limit are inserted into the DynamicFactor set.

const int DynamicFactor_SievingThreshold = MIN(SingleLargePrime_Threshold,100000000);
// up to this value we also actively sieve with SingleLarge-Primes.
// (If the value is smaller than SingleLargePrime_Threshold, then
//  there is a cut-off in sieving functions for dynamische factors.)

#define TSIEVEELEMENTSIZE 1 /* 1 = signed char, 2 = short signed int, 4 = signed int, 0 = no assembler inline code (=generic code) */
// this define triggers some types to be set to approriate values.
// if you don't define it here, a default value will be used, which is perfectly okay.
// if ASM386 is defined, then TSIEVEELEMENTSIZE defines the type which is used in inline-assembler subroutines


class StaticFactorbase
{
public:
  typedef unsigned short int FBsizetype; // sizetype of the maximum size of factorbase
  //typedef int FBsizetype; // sizetype of the maximum size of factorbase

  // If the factorbase should have not more than 65535 elements,
  // you should set FBsizetype to "unsigned short int" to save memory;
  // otherwise "int" is a good choice (thereby reducing implicit type casts).
  //
  // "FBsizetype" has *only* to do with the number of elements in the static
  // factorbase. It has *nothing* to do with the type (and the size) of the
  // prime numbers in the static factorbase itself!

protected:
 static FBsizetype Size_StaticFactorbase;

public:
 static const FBsizetype Size() { return Size_StaticFactorbase; }

 static const FBsizetype MaxSize = 65000; /* maximum size of the static factorbase */
 /* Je grer die zu faktorisierende Zahl ist, desto grer
    sollte prinzipiell auch die Primzahlbasis sein, da dann die Relationen
    einfacher gefunden werden knnen.
    Aber: Das lineare Gleichungssystem sprengt bei groer Primzahlbasis
    den Speicher!
    Fr kleinere zu faktorisierende Zahlen wirkt sich dagegen eine grere
    Primzahlbasis verlangsamend aus, da fr ein greres Gleichungssystem
    natrlich auch mehr Relationen gesiebt werden mssen.
    Je nach Geschwindigkeit des Rechners, des verfgbaren Haupt- und
    Hintergrundspeichers (und auch dessen Zugriffszeiten) knnen die
    ttschlich zu verwendenden Werte stark variieren. Sie sind deshalb
    experimentell zu ermitteln.
    Die Maximalgre wird durch MaxSize_StaticFactorbase festgelegt und steht
    damit bereits zur Kompilierzeit fest.
    Bei einer Speicherkapazitt von 32 MB (realem) Hauptspeicher
    hat sich 20000 als brauchbarer Wert erwiesen.
 */

 static const int max_additional_Powers = 1000; // maximum count of additional powers of primes, with which sieving is done

 friend void tune_parameters(const unsigned int);
};

#endif /*  QSIEVE_FWD_HEADER_ */
