#ifndef qsieve_header
#define qsieve_header

/*! @file
 * @brief
 * main header file of the "Faktorisierungsprogramm"
 */

#include "qsieve-fwd.H"
#include "utils.H"

// use wrapped C++ stream I/O for mpz_t (multiple precision numbers)
#include "mpz_wrapper.H"
using namespace my_mpz_wrapper;

#include "Tfactor.H"
#include "TinyVector.H"
typedef CTinyVector<StaticFactorbase::FBsizetype> CTinyFBsizetypeVector;

#include "myBitString.H"

using std::streampos;

class CProvideHelperVariables : private ForbidAssignment
{
 // it's not nice, but better than providing temporary mpz_t x,y
 // in the global namespace: threadsafe & restricted to the class
 // And we have no need to change the source code at 1000 places...
 protected:
  mutable mpz_t x,y; // two helper variables
 public:
  CProvideHelperVariables() { mpz_init(x); mpz_init(y); }
  ~CProvideHelperVariables() { mpz_clear(x); mpz_clear(y); }
};

class CRelation : protected CProvideHelperVariables
/* CRelation manages a relation of the form
       Delta^2 = product of factors (mod N)
   Our aim is to reduce the set of factors by combining relations.
   If two factors are equal, their product will be a square and can
   thus be eliminated (by taking the factor into Delta).

   If all factors are elininated, the product on the right side will evaluate
   to 1. -> The system of equations is solved.

   Delta will be adjusted dynamically.
*/ 
{
 public:
  static const int no_factor = -1;
  static const int dynamic_factor = -2;
  static const int special_factor = -3;
  int relevant_factor;
   // either the value is one out of { no_factor, dynamic_factor, special_factor }
   // or it is indexing a static factor (in the static factor base)

 private:
  CTinyFBsizetypeVector *Relation_sparse;
  myBitString *Relation_dense;
  mpz_t Delta; // dynamischer Delta-Wert

 public:
  inline CRelation() // constructor
   : CProvideHelperVariables(), relevant_factor(no_factor),
     Relation_sparse(new CTinyFBsizetypeVector), Relation_dense(NULL)
    {
      mpz_init_set_ui(Delta,1);
    }

  inline ~CRelation() // destructor
    {
      if (Relation_sparse) { delete Relation_sparse; Relation_sparse=NULL; }
      if (Relation_dense) { delete Relation_dense; Relation_dense=NULL; }
      mpz_clear(Delta);
      relevant_factor=no_factor;
    }
  
  explicit CRelation(const signed int SievePos, short int HitCount=0); // constructor
  /* initialize "Deltas" with the value of Delta,
     fill in "Relationen" with the prime factors (containing odd exponents)
     of the reduced square.
   */

 public:
  int largest_factor_in_Relation() const
    /* (index) of biggest factor in relation */
    {
      if (Relation_sparse)
	if (Relation_sparse->empty()) return no_factor;
        else 
	  {
	    return Relation_sparse->last();
	  }
      else 
	{
	  return Relation_dense->last(1);
	  /*
	    int h = Relation_dense->last(1);
	    if (h>=0) return h;
	    else return no_factor;
	  */
	}
    }

  int second_largest_factor_in_Relation() const
    /* (Index) of second-largest factor in relation */
    {
      if (Relation_sparse)
        if (Relation_sparse->size()>1)
          return (*Relation_sparse)[Relation_sparse->size()-2];
        else
          return no_factor;
      else // Relation ist dense
        {
          signed int vorletzter_Faktor=Relation_dense->last(1);
          if (vorletzter_Faktor>=0)
           return Relation_dense->prev(vorletzter_Faktor,1);
          else
           return no_factor;
        }
    }

  inline unsigned int SizeOfRelation() const
    {
      if (Relation_sparse) return Relation_sparse->size();
      else return Relation_dense->count(1);
    }

  /*!
   * optimize the size of the relation; if it is useful, then
   * implictly convert the relation between dense and sparse. 
   */
  inline void optisize(void)
    {
      // optimize (minimize) memory consumption of the relation
      // 1. if appropriate, implicitely convert dense<->sparse,
      // 2. minimize memory consuption inside the representation

      const int size=SizeOfRelation();
      const int largest=largest_factor_in_Relation();
      const int quotient = size ? largest/size : 0;

      if (Relation_sparse)
       {
         if (quotient<16-2) convert_Relation_to_dense();
         else Relation_sparse->optisize();
       }
      else // (Relation_dense)
       {
         if (quotient>16+2) convert_Relation_to_sparse();
         else Relation_dense->optisize();
       }
    }
  
  void convert_Relation_to_dense();
  void convert_Relation_to_sparse();

  /// returns, whether relation is empty
  inline bool empty() const
    {
      return (relevant_factor==no_factor);
      // if (Relation_sparse) return Relation_sparse->empty();
      // else return Relation_dense->last(1)==-1; 
    }
	
  void combine(const CRelation& GL2);
  /// compute union with GL2 and remove intersection (-> compute symmetric difference)

  // routines for speeding up multiple combine steps
 protected:
  typedef unsigned short int TExponentArrayElement;
  static TExponentArrayElement ExponentArray[StaticFactorbase::MaxSize+32]; // (Startwert mu {0} sein!)
   // +32 because we want to allow fast routines to avoid costly range checks...
  static unsigned int multi_combine_Counter; // (Startwert mu 0 sein!)
 public:
  void multi_combine_init();
  void multi_combine_main(const CRelation& GL2);
  void multi_combine_exit();
  
  const bool ComputeQuadraticCongruence() const;
  void SanityCheck() const; // check, whether the relation is valid (congruency must be valid)
 private:
  inline void swap(CRelation &GL2)
    { 
      // swap content of this relation with the content of GL2
      // (using reference semantic)
      std::swap(Relation_sparse,GL2.Relation_sparse);
      std::swap(Relation_dense,GL2.Relation_dense);
      std::swap(relevant_factor,GL2.relevant_factor);
      mpz_swap(Delta,GL2.Delta);
    }
 public:
  inline const bool operator< (const CRelation &GL2) const
    {	
      return relevant_factor < GL2.relevant_factor;
    }
  streampos save(ostream &out, const CmpqsFactor factor, const short int HitCount=0) const;
  inline streampos save(ostream &out, const int i=1, const short int HitCount=0) const
   {
     CmpqsFactor DLP;
     DLP=i;
     return save(out,DLP,HitCount);
   };
  CmpqsFactor combine(istream &in, const streampos pos);
  CmpqsFactor multi_combine_main(istream &in, const streampos pos);
  CmpqsFactor combine(istream &in);
  CmpqsFactor multi_combine_main(istream &in);
  friend ostream& operator<< (ostream& ostr, const CRelation& GL);
  friend class StaticRelations;
  friend class SpecialRelations;
  friend class Cprocess_clients;
};


/// stream output function for relations
ostream& operator<< (ostream& ostr, const CRelation& GL);


/*!
 * @short
 * This class handles a double large prime factor for accessing
 * it on file.
 * (special factor = factor containing two (single) large primes)
 */
struct TSpecialFactorRelation {
  CmpqsFactor factor;
  streampos fpos;

  inline bool operator() (const TSpecialFactorRelation &t1, const TSpecialFactorRelation &t2) const
    { 
      return t1.factor<t2.factor;
    }
};


/*!
 * @short
 * handles a (dynamic) single large prime factor for file access
 *
 * This class handles a (single) large prime factor for accessing
 * it on file.
 * (single large prime = prime outside the static factorbase
 *  and inside the dynamic factorbase; if we use the SLP for sieving,
 *  then we call it a dynamic factor.)
 */
class TDynamicFactorRelation
{
  /*
     If factor is not too big, then we can sieve with it as a dynamic factor.
     At the moment dynamic factors are restricted to "int".
  */
public:
  int factor; // factor is given as an integer value
  int SQRT_kN;
  streampos fpos;
  inline bool operator() (const TDynamicFactorRelation &t1, const TDynamicFactorRelation &t2) const
    { 
      return t1.factor<t2.factor;
    }
  inline const bool sieveable() const
   {
     // are we allowed to sieve with this Large Prime?
     return factor<DynamicFactor_SievingThreshold;
   }
};


/*!
 * @short
 * Contains a prime inside the dynamic factorbase and its
 * hits in a specific mpqs sieve interval.
 */
struct TSieve_Delta
{
  int delta;
  int factor;
  inline bool operator< (const TSieve_Delta &v) const
    {
#if defined (USE_FIBHEAP)
      return (delta < v.delta);
#else
      return (delta > v.delta);
#endif
      // delta > v.delta , if queue is used
      // delta < v.delta , if fibonacci-heap is used
    }
};

/// Get a (compressed) value from the stream return it decompressed
int GetValue(istream &in);

/// Put a value (compressed) to the stream
void PutValue(ostream &out, const int w);

#endif
