/* Faktorisierungsprogramm
   written 1997-2000, 2002-2004 by Thorsten Reinecke
   version 2.99: 2004-11 by Thorsten Reinecke
   
   Fr den mathematischen Hintergrund siehe:
   Riesel, Hans: "Prime Numbers and computer methods for factorization",
   2nd ed., Boston; Basel; Stuttgart: Birkhuser, 1994
    
   Das Programm setzt einige der dort beschriebenen Algorithmen (in freier
   Variation) um.
   Zunchst findet eine Probedivision fr kleine Primzahlen statt.
   Anschlieend wird mit der Pollard-rho-Methode nach weiteren kleineren
   Faktoren gesucht.
   Falls um eins verminderte Faktoren ausschlielich in kleinere Teiler
   zerfallen, so werden diese mittels der Pollard-Phi-Methode entdeckt.
   
   Wenn die bis dahin verbleibende Zahl nachweislich zusammengesetzt
   (also nicht pseudoprim) ist, wird eine Faktorisierung durch eine
   Variation des multipolynomialen Quadratischen Siebs versucht.
   
   (vgl. Robert D. Silverman: The Multiple Polynomial Quadratic Sieve,
    Mathematics Of Computation, Vol.48, January 1987, p. 329-339)

   Die dort beschriebene Large Prime Variante verwende ich ebenfalls.
   Zustzlich benutze ich aber Large Primes als dynamische Faktoren zur
   dynamischen Erweiterung der (usrprnglich statischen) Faktorbasis, d.h.
   ich siebe zustzlich mit ihnen. Damit das trotz ihrer Gre effizient
   geschehen kann, benutze ich Priority Queues (z.B. Fibonacci-Heaps).
   Diese Methode ist auf eigenem Mist gewachsen; in der von mir studierten
   Literatur konnte ich keinen Hinweis auf hnliche Methoden finden, die
   die Faktorbasis dynamisch erweitern. -- Zustzlich zu diesem Verfahren
   habe ich mich durch ein als "Double Large Prime"-Variante bekanntes
   Verfahren inspirieren lassen. Dieses ist teilweise implementiert und an
   meine dynamische Erweiterung angepat.
   Eine Zyklensuche zur Konstruktion von vollstndig in der Faktorbasis
   aufgehenden Relationen aus Double Large Primes (LP1,LP2 - LP2,LP3 - ...
   ... - LPk,LP1) ist nunmehr ebenfalls implementiert.
   Weitergehende Verbesserungen fehlen noch:
   Eine Absplittung dynamischer Faktoren aus "triple large primes"
   unter Zuhilfenahme von "double large primes"-Zyklen gerade bei der
   Faktorisierung sehr groer Zahlen drfte z.B. ebenfalls sehr ntzlich
   sein, da sie die dynamische Faktorbasis vergrert...

   Das durch die gesiebten Relationen entstehende lineare Gleichungssystem
   wird dabei simultan mit den eingefgten Relationen gelst.
   Um Speicherplatz zu sparen, die Korrektheit bertragener Relationen beim
   verteilten Sieben einfacher berprfen zu knnen sowie triviale Lsungen
   schneller zu entdecken, habe ich die Relationen normiert:
   Jede Relation hat bei mir nun die Form 
    "enthaltene Faktorbasiselemente mit ungeradem Exponenten" * x^2 = 1 (mod n),
   wobei x den bereits ermittelbaren Quadratwurzelanteil darstellt.
   Mehrfachpotenzen der enthaltenen Faktorbasiselemente werden bei
   (ganzzahliger) Halbierung des Exponenten zu x hinzumultipliziert. -- Wenn
   schlielich eine Relation der Form x^2=1 entstanden ist, so kann ein
   (hoffentlich nichttrivialer) Faktor ber gcd(x-1,n) gefunden werden.

   Es ist auch eine Implementierung elliptischer Kurven zur Ermittlung von
   Faktoren "gemigter Gre" enthalten. Dies ist das einzige Verfahren,
   mit dem sich momentan auch bei sehr groen Zahlen nichttriviale Faktoren
   entdecken lassen (da seine Laufzeit von der Gre des entdeckten Faktors
   abhngig ist).

   In seiner jetzigen Fassung sollte man mit dem Programm
   Zahlen bis um 110 Dezimalstellen in ertrglicher Zeit faktorisieren knnen.
  
   [ 
     Nachtrag Juni 2003:
     Bei greren Zahlen lassen sich mit den elliptischen Kurven immerhin
     nichttriviale Faktoren absplitten, z.B. als Dauerproze im Hintergrund:
     P190169 (480 Dezimalstellen); entdeckter Faktor:
     9820062895352727932801465820946503529 (37) [ecm,sigma=272319096]

     Nachtrag September 2003:
     Im kleinen Netzwerk bentigte die Version 2.92 fr die Faktorisierung einer
     107-stelligen Lucaszahl mit dem Siebverfahren ca. 5 Wochen Rechenzeit.
     (Die Faktorisierung wurde dabei als reiner Hintergrundprozess betrieben.)

     Nachtrag August 2004:
     Die Version 2.98 bentigte (unter hnlichen Bedingungen) fr verschiedene
     107-stellige Zahlen jeweils zwischen 8 und 14 Tage.
   ]
   
   Es fehlt allerdings noch eine ganze Reihe an Optimierungen und Algorithmen.
   Fr ein effizienteres Sieben wre vielleicht noch das "Number-Field-Sieve"
   sinnvoll. (Siebverfahren faktorisieren grundstzlich in Abhngigkeit von
   der Gre der zu faktorisierenden Zahl; sie sind besonders effizient,
   wenn die Zahl lediglich zwei Primteiler hat und diese von der Stelligkeit
   her nicht allzuweit voneinander entfernt liegen.)

   Ob das simultane Lsen des linearen Gleichungssystems sinnvoll ist, oder
   ob man vielleicht effizientere Methoden (z.B. Wiedemann-Verfahren) nach
   Abschlu der Siebphase anwendet, hngt im wesentlichen vom Speicherbedarf
   des Gleichungssystems ab. Beim verteilten Sieben hat das simultane Lsen
   des Gleichungssystems zumindest den Vorteil, parallel zur Siebphase abzu-
   laufen. 
*/

/*! @file
 * @brief
 * main implementation file of "Faktorisierungsprogramm"
 */


#ifdef AUTODETECTION
 // autodetection for gcc-compiler
 #warning "autodetection enabled..."
 #ifdef __GNUG__
  #if __GNUC__ > 2
  #else
    #error "Please compile again with a newer version of GCC (version >= 3.2.x, better >=3.4.x)"
  #endif

  #if #cpu (i386)
   #ifndef ASM386
    #warning "activating i386-inline assembler code"
    #define ASM386
    #ifdef ASMATHLON
     #warning "using athlon-inline assembler code"
     #ifndef ASM386_CMOV
     #warning "activating ASM386-CMOV usage, because ATHLON supports it :)"
     #define ASM386_CMOV
     #endif
    #endif
   #endif
  #else
   #ifdef ASM386
    #warning "deactivating i386-inline assembler code"
    #undef ASM386
    #undef ASM386_CMOV
    #undef ASMATHLON
   #endif
  #endif
  #if !defined (unix) && defined(USE_NETWORK)
    #error "network support only for Unix-systems" 
  #endif
 #endif
#endif /* autodetection */

using namespace std;


#ifndef USE_NETWORK
 #ifdef IS_SERVER
  #error "Illegal configuration: SERVER and no USE_NETWORK"
 #endif
#else
 #ifndef IS_STANDALONE
  #ifdef IS_SERVER
   #include <map>
  #endif
  #include "unix_buffer.H"
 #endif
#endif

#ifndef IS_SERVER
 #ifndef IS_CLIENT
  #undef IS_STANDALONE
  #define IS_STANDALONE
 #endif
#endif

#ifdef IS_SERVER
 #ifdef IS_CLIENT
  #error "Illegal Configuration SERVER-CLIENT"
 #else
  #ifdef IS_STANDALONE
   #error "Illegal Configuration SERVER-STANDALONE"
  #endif
 #endif
#endif

#ifdef IS_CLIENT
 #ifdef IS_STANDALONE
  #error "Illegal Configuration CLIENT-STANDALONE"
 #endif
#endif

#if defined(USE_FIBHEAP) && defined(USE_FAKEHEAP)
 #error "defines USE_FIBHEAP and USE_FAKEHEAP exclude each other!"
#endif

#ifdef REACT_ON_SIGUSR
 // a little class that catches signal SIGUSR1 and SIGUSR2
 // and stores these flags...
 // can be useful to abort some long-running algorithms
 // without aborting the program or altering configuration parameters.
 #include "usr_signals.cc"
 Cusr_signal_proxy USRSignalHandler;
#endif


// propagate IS_STANDALONE-condition from compiletime to runtime,
// so that certain modules can be compiled independently of it.
const bool compiled_IS_STANDALONE()
{
  #ifdef IS_STANDALONE
   return true;
  #else
   return false;
  #endif
}


extern "C"
{
 #include <unistd.h>
}

#include <cstdio>
#include <cstdlib>
#include <string>
#include <iostream>
#include <iomanip>
#include <fstream>
#include <sstream>
#include <cmath>
#include <gmp.h>
#include <list>
#include <set>
#include "utils.H"
#include <ctime>

#include "qsieve.H"


struct myintcompare
{
  inline const bool operator() (const int i1, const int i2) const
   {
     return i1 < i2;
   }
};
typedef set<int, myintcompare> intset;


class TFoundFactor
{
 private:
  mpz_t x;
  unsigned int e;
  string comment;
 public:
  explicit TFoundFactor(const mpz_t f, const unsigned int exp = 1, const string &xcomment="")
   : e(exp), comment(xcomment) { mpz_init_set(x,f); }
  explicit TFoundFactor(const long unsigned int f, const unsigned int exp = 1)
   : e(exp), comment() { mpz_init_set_ui(x,f); }
  TFoundFactor(const TFoundFactor &FF)
   : e(FF.e), comment(FF.comment) { mpz_init_set(x,FF.x); }
  explicit TFoundFactor(std::istream &is)
   {
     mpz_init(x);
     is.get();
     if (is.eof()) { is.setstate(std::ios::failbit); return; }
     is.unget();
     string s;
     is >> x;
     is >> s; if (s!="e") { MARK; is.setstate(std::ios::failbit); return; }
     is >> e;
     is >> s; if (s!="c") { MARK; is.setstate(std::ios::failbit); return; }
     char buf[4096];
     is.getline(buf,sizeof(buf),'\n'); comment=buf;
   }
  virtual ~TFoundFactor() { mpz_clear(x); }
  virtual bool operator< (const TFoundFactor &nf) const { return (mpz_cmp(x,nf.x)<0); }
  virtual void output(std::ostream &ostr) const;
  TFoundFactor& operator= (const TFoundFactor &rhs)
   {
     mpz_set(x,rhs.x); e=rhs.e; comment=rhs.comment;
     return *this;
   }
  virtual void Save(std::ostream &os) const
   {
     os << x << " e " << e << " c " << comment << endl;
   }
};



inline std::ostream& operator << (std::ostream& ostr, const TFoundFactor FF)
{
  FF.output(ostr);
  return ostr;
}

// quick and dirty:
class TFoundFactors : public set<TFoundFactor>
{
public:
 string regarding; // argument (given number, can be symbolic)

 void Load(std::istream &is)
  {
    clear();
    while (is)
     {
       TFoundFactor FF(is);
       if (is.good()) insert(FF);
     }
  }
 void Save(std::ostream &os) const
  {
    for (TFoundFactors::const_iterator p=begin(); p!=end(); ++p) p->Save(os);
  }

 void AutoSave() const;
 void AutoLoad();
};

TFoundFactors FoundFactors;


#include <vector>
#include <stack>
#include <algorithm>

#ifdef USE_NETWORK
 extern "C"
 {
  #include <sys/utsname.h>
  #include <pthread.h>
 }
 #include <csignal>
 #include "mutex.H"
 #ifdef IS_CLIENT
  extern "C"
  {
   #include <sys/wait.h>
   #include <unistd.h>
  }
 #endif
#endif

#ifdef NOTIFY_PARENT
#if defined(IS_SERVER) && defined(USE_NETWORK) 
 #include <csignal>
#else
 #warning "undefining NOTIFY_PARENT, since this is not a networked server"
 #undef NOTIFY_PARENT
#endif
#endif


// output to screen/terminal can optionally be handled by a ncurses based
// wrapper, but on default all windows will be merged into cout.
#ifdef USE_NCURSES
 #include <my_ncurses.H>
#endif

ostream cout_titel(cout.rdbuf());
ostream cout_network(cout.rdbuf());
ostream cout_status(cout.rdbuf()); 


const string RecoveryFile         = "recovery.dat";
const string FoundFactorsFile     = "FoundFactors.dat";
const string FactorizationsFile   = "factorizations.txt";
const string StaticRelationsFile  = "static_relations.dat";
const string SpecialRelationsFile = "special_relations.dat";
      string DynamicRelationsFile = "dynamic_relations.dat"; // default, may be overwritten by entry in configfile
      string TempDir              = "/tmp"; // default directory for temporary files
      string ConfigFile           = "qsieve.cfg"; // default, may be overwritten by getenv("QSIEVE_CFG")


// provide streams for storing relations and other data
// *************************************************************************
// **** important:
// **** If "ostream" and "istream" are handled by the same filebuf,
// **** then you *must not* rely on 
// **** "tellp()" and "tellg()" behave independently from each other!
// **** Any write operation may change also the value of "tellg()"
// **** and any read operation may change the value of "tellp()"!
// ************************************************************************* 

// provide streams to store SLP relations
// (SLP relations = SingleLargePrime relations;
//  synonymic: dynamic factor relations, dynamic relations)
filebuf DynamicRelations_buffer;
ostream DynamicRelations_to_file (&DynamicRelations_buffer);
istream DynamicRelations_from_file (&DynamicRelations_buffer);

// for Recovery
filebuf Recovery_buffer;
ostream Recovery_to_file (&Recovery_buffer);
istream Recovery_from_file (&Recovery_buffer);

#ifdef IS_CLIENT
string communication_name;
ofstream communication_stream;
#endif

#ifndef IS_CLIENT
std::ofstream Factorization_to_file;
/* Factorizations are logged to this file. Numbers to factorize and any
   found factors are appended. The file can be deleted, if the logged data
   is of no use anymore.
*/
#endif

#include "mpqsPolynom.H"


#ifndef IS_SERVER
#ifdef USE_FIBHEAP
 #include "fibheap.H"
 FibHeap <TSieve_Delta> Sieve_Delta_Heap, Sieve_Delta_HeapOfSquares;
 #include <queue>
#elif defined(USE_FAKEHEAP)
 #include "fakeheap.H"
 FakeHeap <TSieve_Delta> Sieve_Delta_Heap;
 #include <queue>
 priority_queue<TSieve_Delta> Sieve_Delta_HeapOfSquares;    
#else
 #include <queue>
 priority_queue<TSieve_Delta> Sieve_Delta_Heap, Sieve_Delta_HeapOfSquares;
#endif
#endif

mpz_t n, // number to factorize (will be reduced during factorization)
      kN; // input for MPQS (includes a suitable multiplier)

int Vorfaktor_von_kN; // pre-multiplier for kN (kN=Vorfaktor_von_kN*n), will be determined later!


int PrimeNumbers[StaticFactorbase::MaxSize]; /* fr eine effiziente Probedivision
				       alle Primzahlen der Faktorbasis hier
				       akkumulieren */


int SQRT_kN_of_PrimeNumbers[StaticFactorbase::MaxSize];
int Delta_of_PrimeNumbers[StaticFactorbase::MaxSize][2];
    /* for efficient access the values (per sieve interval) are placed here. */
 
int NumberOf_more_PrimePowers = 0; // actual number (will be evaluated in runtime)
int FB_maxQuadrate = 0; // actual number of squares to sieve with (will be evaluated in runtime)
int PrimePowers[StaticFactorbase::max_additional_Powers];
int SQRT_kN_of_PrimePowers[StaticFactorbase::max_additional_Powers];
int Delta_of_PrimePowers[StaticFactorbase::max_additional_Powers][2];
int SQRT_kN_of_PrimeSquares[StaticFactorbase::MaxSize];

#include "StaticRelations.H"


/* About Dynamic-Factors/Special-Factors:

   This program uses a modified Multi-Large-Prime-Variant:
   Dynamic-Factors are like Single-Large-Primes, except that we are
   sieving with these factors as they get detected.

   Special-Factors are composite rests (esp. Double-Large-Primes-Variante). 
   On the one hand, we try to split Special-Factors by new Dynamic-Factors
   to get other Dynamic-Factors.  On the other hand a cycle search is
   performed occasionally to gain relations.

   In future implementations the cycle search may be extended to hyper edges
   as well (for multi-Large-Primes). [But most probably this approach would
   end up in solving a linear system of equations.]
*/

typedef set<TDynamicFactorRelation,TDynamicFactorRelation> TDynamicFactorRelations;

#ifdef IS_SERVER
class CServerDynamicFactorRelations : public TDynamicFactorRelations
{
 private:
  mutable CMutex monitor_mutex;
  vector<int> DynamicFactorList;
  int size_active_SLP, size_passive_SLP;
 public:
  inline void insert(const TDynamicFactorRelation& x)
   {
     if (TDynamicFactorRelations::insert(x).second)
      {
        if (!x.sieveable())
         {
           // x is a Single Large Prime, but a passive one: not sieveable
           ++size_passive_SLP;
           return;
         }
        monitor_mutex.lock();
        ++size_active_SLP;
        DynamicFactorList.push_back(x.factor);
        monitor_mutex.unlock();
      }
     else
      {
        cerr << "CServerDynamicFactorRelations::insert duplicate (rejected)!" << endl;
        exit(1);
      }
   }
  inline const int monitoredSize() // const /* const attribute could be dangerous due to multithreading!? */
   {
     monitor_mutex.lock();
     const int s=DynamicFactorList.size();
     monitor_mutex.unlock();
     return s;
   }
  inline const int size_active() const { return size_active_SLP; }
  inline const int size_passive() const { return size_passive_SLP; }
  inline const int operator[] (const int i) // const /* const attribute could be dangerous due to multithreading!? */
   {
     monitor_mutex.lock();
     const int x=DynamicFactorList[i];
     monitor_mutex.unlock();
     return x;
   }
};
CServerDynamicFactorRelations DynamicFactorRelations;
#else
TDynamicFactorRelations DynamicFactorRelations;
#endif

double Factor_Threshold = 3.0; /* exponent, which allows bigger factors (which are not part of the static factorbase)
				  suggestive values: 1.0 up to 3.0
				  1.0: no additional factors outside the static factorbase
				  up to 2.0: prime factors outside the static factorbase
                                             up to the square of the biggest factor inside the static FB.
				  >2.0: factors up to (biggest factor inside the static FB) raised to the power of Factor_Threshold.
			       */


#include "SpecialRelations.H"


bool collecting_phase_finished = false;
/* switch to define the "active" and "passive" phase of sieving:
 If enough Special-Relations and Dynamic-Relations have been collected to
 sieve Static-Relations at a high rate, then further collecting even more
 Special- and Dynamic-Relations can be suppressed by setting
 collecting_phase_finished=true.
*/

int biggest_Prime_in_Factorbase;


#include "Sieving.H"
#ifdef IS_SERVER
 // we have to define some variables, because
 // Sieving.o is not linked to the binary for servers...
 int LogicalSieveSize = 1000000; // sieving interval will be [-LogicalSieveSize,LogicalSieveSize] for each MPQS polynomial
#endif


// a few parameters for easy-factorization (and their default values)
bool SkipFermat    = false; // try the fermat method?
bool SkipPhi       = false; // try Pollard-Phi-Method?
bool SkipFibonacci = false; // try Fibonacci-Method?
bool SkipEasyECM   = false; // try elliptic curve method im easy-mode?
bool SkipECM       = false; // try elliptic curve method?
int rho_Phase      =   150000; // main initialization for Rho-Phase
int phi_Phase1     =   500000; // upper bound for Phase 1 in phi-Method
double phi_Phase2  = 10000000.0; // upper bound for phi-Method Phase 2 (=isolated factor)
int elcu_Phase1    =    20000; // upper bound for elliptic curve Phase 1
double elcu_Phase2 =  2000000.0; // upper bound for elliptic curve Phase 2
int elcu_Kurven    =      100; // number of elliptic curves to try
int ecm_curves_processed = 0; // number of curves done so far


CmpqsPolynom Polynom; // object to manage polynimal computations fr multipolynomial sieve (MPQS)
#include "mpqsStatistics.cc" // statistical stuff about found relations, client connections, etc.


inline bool is_dynamic_factor(const int number)
{
  // returns, whether this number is already registered as a dynamic factor;
  // (dynamic factor -> prime number outside the static factorbase, also known as Single Large Prime)
  TDynamicFactorRelation relation;
  relation.factor = number;
  const TDynamicFactorRelations::iterator pos = DynamicFactorRelations.find(relation); // search dynamic factor
  return (pos != DynamicFactorRelations.end());
}


#include "modulo.H" // modulo operations for unsigned int
using namespace numtheory;


const int probab_prime_checks = 25; // #rounds for mpz_probab_prime

string XML_StatusFile;


void TFoundFactor::output(std::ostream &ostr) const
{
  ostr << x;
  if (e!=1) ostr << "^" << e;
  if (comment!="") ostr << " " << comment;
  else
   {
     if (!mpz_probab_prime_p(x,probab_prime_checks)) ostr << "[composite]";
   }
}

void TFoundFactors::AutoSave() const
{
  std::ofstream ofstr(FoundFactorsFile.c_str());
  ofstr << regarding << endl;
  Save(ofstr);
}

void TFoundFactors::AutoLoad()
{
#ifdef VERBOSE_INFO
  cout << "Autoload FoundFactors" << endl;
#endif
  std::ifstream ifstr(FoundFactorsFile.c_str());
  char buf [32768];
  ifstr.getline(buf,sizeof(buf),'\n'); regarding=buf;
  Load(ifstr);
}


#include "ConfigFile.cc"


const char* const _MAL(void)
{
  // returns empty string on first call, and  " * " otherwise.
  // Task: Prevent costly checks for suppressing the trivial factor 1
  // as the fist resp. last factor.

#ifdef NOTIFY_PARENT
  // send SIGUSR2 to the parent to notify that a new
  // factor has found
  kill(getppid(),SIGUSR2); // send SIGUSR2 to the parent process
#endif

  static bool first_call = true;
  if (first_call)
   {
     first_call=false;
     return "";
   }
  else return " * ";
}

std::string MAL(const mpz_t factor, const unsigned int exponent=1)
{
  FoundFactors.insert(TFoundFactor(factor,exponent));
  FoundFactors.AutoSave();
  ostringstream helper;
  helper << _MAL() << factor;
  if (exponent>1) helper << "^" << exponent;
  return helper.str();
}

std::string MAL(const mpz_t factor, const unsigned int exponent, const std::ostringstream &comment)
{
  FoundFactors.insert(TFoundFactor(factor,exponent,comment.str()));
  FoundFactors.AutoSave();
  ostringstream helper;
  helper << _MAL() << factor;
  if (exponent>1) helper << "^" << exponent;
  return helper.str()+comment.str();
}


std::string MAL(const mpz_t factor, const std::ostringstream &comment)
{
  FoundFactors.insert(TFoundFactor(factor,1,comment.str()));
  FoundFactors.AutoSave();
  ostringstream helper;
  helper << _MAL() << factor;
  return helper.str()+comment.str();
}

std::string MAL(const long unsigned int factor, const unsigned int exponent=1)
{
  FoundFactors.insert(TFoundFactor(factor,exponent));
  FoundFactors.AutoSave();
  ostringstream helper;
  helper << _MAL() << factor;
  if (exponent>1) helper << "^" << exponent;
  return helper.str();
}


#include "polynomial.H" /* for fast polynomial arithmetic & fft-continuation */
#include "fft_param.cc"  /* discrete fast fourier transform */
#include "elliptic_curve.H" /* invariant part of elliptic curve method (ecm) */
#include "elliptic_curve-variant.cc" /* variant part of elliptic curve method (ecm) */

#include "ExitManager.cc" /* controlled termination of program */
#include "StaticRelations.cc"

#ifndef IS_CLIENT
 // this stuff may only be needed by server or stand-alone version
 #include "easy_factor.H"
#endif /* of #ifndef IS_CLIENT */


#include "sqrt_modulo.cc"

inline int SQRT_kN_mod_PrimeNumber(const unsigned int Primzahl)
{
  return sqrtmod(mpz_remainder_ui(kN,Primzahl),Primzahl);
}


int check_SQRT_kN_mod_PrimeNumber(const int Primzahl)
{
  // can be called for debugging
  // (and to check, whether "SQRT_kN_mod_PrimeNumber" works correctly)
  int w1 = SQRT_kN_mod_PrimeNumber(Primzahl);
  if (w1>Primzahl/2) w1=Primzahl-w1;

  mpz_t x;
  mpz_init_set_ui(x,w1); mpz_mul(x,x,x); mpz_sub(x,x,kN);
  if (mpz_mod_ui(x,x,Primzahl)!=0)
    {
      MARK;
      cerr << "wrong result! " << w1 << " mod " << Primzahl << endl;
      exit(1);
    }
  mpz_clear(x);
  return w1;
}


void compute_StaticFactorbase()
{
#if defined(ASMATHLON) || defined(ASM_P3_SSE)
  // ATHLON specific sanity checks
  if ( LogicalSieveSize>(1<<23) || (StaticFactorbase::Size()&3) )
   {
     MARK;
     cerr << "Sorry! Processor specific constraints not met." << endl;
     cerr << "For using AMD-3DNow! or SSE optimized version we need:" << endl;
     cerr << "Constraint 1 (SIMD): (Size_StaticFactorbase&3)==0" << endl;
     cerr << "Constraint 2 (float precision): LogicalSieveSize<=(1<<23) [=8388608]" << endl;
     cerr << "Please edit ConfigFile according to these constraints" << endl;
     cerr << "OR use a version without these optimizations." << endl;
     exit(1);
   }
#endif

#ifdef VERBOSE_NOTICE
  cout << "Calculating static factorbase." << endl;
#endif
  PrimeNumbers[0]=-1; // for handling signess of the relation
  SieveControl::log_PrimeNumbers_dirty[0]=SieveControl::log_PrimeNumbers[0]=0; // sign has no logarithmic value

  int Primzahl=2;
  // in this implementation of MPQS the prime number 2 is a member of any
  // static factorbase (without exception!)
  PrimeNumbers[1]=2;
  SieveControl::log_PrimeNumbers_dirty[1]=SieveControl::log_PrimeNumbers[1]=static_cast<TSieveElement>(ceil(log_SieveEntryMultiplier*log(Primzahl)));

  Primzahl=1; // initial value (for the loop, expected step: +2)
  mpz_t x,y,wuqu;
  mpz_init(x); mpz_init(y); mpz_init(wuqu); // for computing qsuare root of prime powers
  for (int Primzahl_nr=2; Primzahl_nr<StaticFactorbase::Size(); Primzahl_nr++)
    {
      do // compute next prime number in static factorbase
	{
	  do Primzahl+=2; while(!is_prime(Primzahl)); // next prime number
	  mpz_set_ui(x,Primzahl);
	}
      while (mpz_legendre(kN,x)<0); // until prime number is member of factorbase
      
      PrimeNumbers[Primzahl_nr]=Primzahl;
      
      // needed for efficient sieving (spares repeated recomputing)
      SieveControl::log_PrimeNumbers[Primzahl_nr]=static_cast<TSieveElement>(ceil(log_SieveEntryMultiplier*log(Primzahl)));
      SieveControl::log_PrimeNumbers_dirty[Primzahl_nr]=static_cast<TSieveElement>(ceil(log_SieveEntryMultiplier*log(Primzahl)*SieveControl::DirtyFactor(Primzahl_nr)));
      
      // needed to compute the Delta values for sieving
      SQRT_kN_of_PrimeNumbers[Primzahl_nr]=SQRT_kN_mod_PrimeNumber(Primzahl);
      
      // check, whether computations of square roots were correct:
      if ( squaremod(SQRT_kN_of_PrimeNumbers[Primzahl_nr],Primzahl) != mpz_remainder_ui(kN,Primzahl) )
       {
         cerr << "square root not correct!" << endl;
         cerr << SQRT_kN_of_PrimeNumbers[Primzahl_nr] << "^2 <> "
              << mpz_remainder_ui(kN,Primzahl) << " (mod " << Primzahl << ")" << endl;
         exit(1);
       }


      if ( Primzahl < sqrt(std::numeric_limits<int>::max()) && Vorfaktor_von_kN%Primzahl!=0 )
       {
          // *** compute square roots of kN mod Primzahl^2 ***
          const unsigned int P = Primzahl*Primzahl;
	  const unsigned int Wu = SQRT_kN_of_PrimeNumbers[Primzahl_nr];
          unsigned int iy = invmod(2*Wu,P);
          unsigned int ix = mpz_remainder_ui(kN,P);
          
	  unsigned int wuq=Wu*Wu;
          if (ix>=wuq) ix-=wuq; else ix=ix-wuq+P; // ix-=wuq; avoid overflow; we compute values modulo P
          ix=mulmod(ix,iy,P)+Wu;
          wuq = (ix<P) ? ix : ix-P; // this is a faster version for: wuq=ix%P

#if 1 || defined(DEBUG)
          if (mulmod(wuq,wuq,P)!=mpz_remainder_ui(kN,P))
           {
             cerr << "PQ-root incorrect!" << endl;
             cerr << "Primzahl=" << PrimeNumbers[Primzahl_nr] << endl;
             cerr << "kN= " << kN << endl;
	     cerr << mulmod(Wu,Wu,P) << " != " << mpz_remainder_ui(kN,P) << endl;
             exit(1); 
           }
#endif
          SQRT_kN_of_PrimeSquares[Primzahl_nr]=wuq;
       } else SQRT_kN_of_PrimeSquares[Primzahl_nr]=-1;


      const int Threshold = LogicalSieveSize < std::numeric_limits<int>::max()/8 ? 8*LogicalSieveSize/Primzahl : LogicalSieveSize/Primzahl;
      if (Primzahl<=Threshold && Primzahl>2 && mpz_remainder_ui(kN,Primzahl)!=0)
	// constraint for Primzahl to assure correct sieving with prime-powers: 
        //  - not too big, bigger than 2, and no divisior of kN
	{
	  FB_maxQuadrate=Primzahl_nr; // at the end of the loop this will be the limit, up to which we can sieve squares the normal way...

	  // compute square roots of kN modulo Primzahl^pot
	  // initial values:
	  int P_Potenz = Primzahl*Primzahl;
	  int Wu       = SQRT_kN_of_PrimeNumbers[Primzahl_nr];

          double Weight = Primzahl_nr<SieveControl::FBStartIndex() ? 2.0 : 1.0;
           // We do not sieve with the first couple of primes of the
           // factorbase (because these primes are very small and sieving
           // with them takes too long); as a minor correction of this
           // "dirty behaviour" the value of the smallest power of each dirty prime
           // can be adapted for sieving: log(P_Potenz)=Potenz*log(Primzahl).
           // All other powers are handled the usual way: since it is a recurrence of a known factor,
           // only log(Primzahl) has to be subtracted.

	  for (int pot=2; ;pot++)
	    {
	      //cout << "computing square root of " << Primzahl << "^" << pot << endl;

	      mpz_set_ui(y,2*Wu); mpz_set_ui(x,P_Potenz);
	      if (mpz_invert(y,y,x)==0)
		{ 
		  cerr << "PPot-root for " << Primzahl  << ": inverse doesn't exist!" << endl;
		  break;
		}

              if (NumberOf_more_PrimePowers>=StaticFactorbase::max_additional_Powers)
               { 
                 static char weitermachen = 'n';
                 if (weitermachen!='y')
                  {
	           cerr << "Overflow for " << Primzahl << "^" << pot << endl;
                   cerr << "Es kann nicht optimal mit Primzahlpotenzen gesiebt werden," << endl
                        << "da der hierfr reservierte Speicherplatz nicht ausreicht." << endl;
                   cerr << "Sie sollten \"StaticFactorbase::max_additional_Powers\" vergroessern!" << endl;
                   cerr << "Reserved memory for storing prime powers is too small..." << endl;
                   cerr << "sieving will be less efficient..." << endl;
                   cerr << "continue? (y/n) " << flush;
                   cin >> weitermachen;
                   if (weitermachen=='n') exit(1);
                  }
                 break; // avaoid overflow by breaking the loop
               }

	      mpz_mod_ui(x,kN,P_Potenz); mpz_set_ui(wuqu,Wu); mpz_mul_ui(wuqu,wuqu,Wu); mpz_sub(x,x,wuqu);
	      mpz_mul(x,x,y); mpz_add_ui(x,x,Wu); mpz_mod_ui(x,x,P_Potenz);
	      Wu = mpz_get_ui(x); if (Wu>P_Potenz-Wu) Wu=P_Potenz-Wu; // normalize to the "lower" of the two square roots
              // "Wu" is now the square root of kN (mod P_Potenz)

	      // check, whether computation was correct:
	      mpz_set_ui(x,Wu); mpz_mul(x,x,x); mpz_sub(x,kN,x); 
	      if (mpz_mod_ui(x,x,P_Potenz)!=0)
		{
		  cerr << "PPot-root incorrect!" << endl;
		  cerr << "Primzahl=" << PrimeNumbers[Primzahl_nr] << " Nr. " << Primzahl_nr << endl;
		  cerr << "kN= " << kN << endl;
		  cerr << "SQRT(kN) mod p=" << SQRT_kN_of_PrimeNumbers[Primzahl_nr] << endl;
		  cerr << "SQRT(kN) mod p^i=" << SQRT_kN_of_PrimePowers[NumberOf_more_PrimePowers] << endl;
		  cerr << "but we got a wrong value: " << x << endl;
		  exit(1);
		}

              // heuristics: Increase the weight of the last power a little bit,
              // since we have a slight chance, that this hit is a hit of higher powers, too.
              if (P_Potenz>Threshold) Weight+=4.0/Primzahl;

#if 1
              // heuristics: (determined by trial&error):
              // For factorizing "small" numbers: Sieving is significantly
              // faster, if sieving is skipped for smaller values;
              // the sieve gets more wide-meshed, but this doen't matter, since relations come plentiful...
              // For factorizing "bigger" numbers: If we sieve more carefully, we detect more DLP in the same time,
              // this effect is by far more important!
              if ( P_Potenz<250 && mpz_sizeinbase(kN,10)<70 ) Weight+=1.0;
              else
#endif
               {
	         PrimePowers[NumberOf_more_PrimePowers]=P_Potenz;
	         SQRT_kN_of_PrimePowers[NumberOf_more_PrimePowers]=Wu;

		 log_Primzahl_of_PrimePowers[NumberOf_more_PrimePowers]
                   =static_cast<TSieveElement>(ceil(log_SieveEntryMultiplier*Weight*log(Primzahl)));
                 Weight=1.0;
                 ++NumberOf_more_PrimePowers; // we got one more...
               }
	      
	      if (P_Potenz>Threshold) break; // next one would be too big...
	      P_Potenz*=Primzahl;
	    }
	}
    }
  mpz_clear(x); mpz_clear(y); mpz_clear(wuqu);

  biggest_Prime_in_Factorbase=Primzahl;
#ifdef VERBOSE_INFO
  cout << "prime numbers in static factorbase have been computed." << endl;
  if (Primzahl>=PhysicalSieveSize)
    cout << "Remark: Some members of the static factorbase are bigger than the sieve size!" << endl;
#endif
}


#ifndef IS_SERVER
void compute_Deltas_for(const TDynamicFactorRelation &DynRel, const int offset)
{
  TSieve_Delta sieb_delta;
  int d1,d2;
  
  sieb_delta.factor=DynRel.factor;

  const unsigned int inv_A2 = invmod(mpz_remainder_ui(Polynom.get_A2(),DynRel.factor),DynRel.factor);
  const unsigned int PolyB = mpz_remainder_ui(Polynom.get_B(),DynRel.factor);

  d1=DynRel.SQRT_kN-PolyB; if (d1<0) d1+=DynRel.factor;
  d1=mulmod(d1,inv_A2,DynRel.factor);
  sieb_delta.delta=normalized_signed_mod(d1-offset,DynRel.factor);
  sieb_delta.delta+=offset;
#ifdef USE_FAKEHEAP
  while (sieb_delta.delta<=LogicalSieveSize) // shortcut: only re-insert, if still inside the sieve interval!
   {
     Sieve_Delta_Heap.push(sieb_delta);
     sieb_delta.delta+=sieb_delta.factor;
   }
#else
  if (sieb_delta.delta<=LogicalSieveSize) // shortcut: only re-insert, if still inside the sieve interval!
    Sieve_Delta_Heap.push(sieb_delta);
#endif
  
  d2=DynRel.factor-DynRel.SQRT_kN-PolyB; if (d2<0) d2+=DynRel.factor;
  d2=mulmod(d2,inv_A2,DynRel.factor);
  sieb_delta.delta=normalized_signed_mod(d2-offset,DynRel.factor);
  sieb_delta.delta+=offset;
  if ( (sieb_delta.delta<=LogicalSieveSize)  && (d1!=d2) ) // shortcut: only re-insert, if still inside the sieve interval!
   {
     Sieve_Delta_Heap.push(sieb_delta);
#ifdef USE_FAKEHEAP
     sieb_delta.delta+=sieb_delta.factor;
     while (sieb_delta.delta<=LogicalSieveSize)
      {
        Sieve_Delta_Heap.push(sieb_delta);
        sieb_delta.delta+=sieb_delta.factor;
      }
#endif
   }
}
#endif




// this is used to store the sieved dynamic factors.
// Remark: To collect the dynamic factors efficiently in the sieving phase, I decided to make this array global.
//         (The constructor of "CRelation" is called with the number of found dynamic factors for each relation.)
const short int MaxHits=64; // "64 dynamic factors should be enough for everyone"
static class
        {
         public: int Faktor;
         private: int Exponent;
         public:
          inline const int GetExponent() const { return Exponent; }
         friend CRelation::CRelation(const signed int SievePos, short int HitCount /* =0 */);
        } Hits[MaxHits]; // collected dynamic factors

// important: The exponent need not to (and shouldn't) be set during the sieving phase. It will be evaluated
// inside the constructor (and only there!) while the relation is splitted into its factors.
// Also keep in mind not to place any composite numbers as factors into this struct.


#ifndef IS_SERVER

CRelation::CRelation(const signed int SievePos, short int HitCount /* =0 */)
: CProvideHelperVariables(), relevant_factor(no_factor), Relation_sparse(new CTinyFBsizetypeVector), Relation_dense(NULL)
  /* constructor: insert a new relation.
     initialize "Deltas" with the value of Delta,
     (re)split the relation into its factors; thereby pick up only the factors, whose exponents are odd
     (if the exponent is even, then the factor can be replaced by factor^2 and thereby be handled as a square!)
   */
{
  // cout << "initialize relation" << endl;
  mpz_init(Delta);
  Polynom.get_values(SievePos,Delta,y);
  // in y: reduced square (which is now to be factorized...) 
  mpz_invert(Delta,Delta,n); /* we use the modular inverse of Delta,
			        this saves us an extra variable;
				doing so, the quadratic congruence
				is normalized to Delta^2=1 (mod n)
			      */
  
  if (mpz_sgn(y)<0) // negative numbers (handle sign)
    {
      Relation_sparse->append(0);
       // -1 must be considered to handle the signess;
       // we have to insert a "0", because PrimeNumbers[0]=-1.
      mpz_neg(y,y); // switch sign (to make positive)
    }
  
  // special handling for the first factor of the static factorbase.
  // In MPQS we have problems to sieve for prime number 2...

  if (PrimeNumbers[1]==2)
   {
     // special treatment for 2, since it it very easy:
     const unsigned int z = mpz_scan1(y,0);
     if (z)
      {
        mpz_tdiv_q_2exp(y,y,z);
        mpz_mul_2exp(Delta,Delta,z>>1);
        if (z&1) Relation_sparse->append(1); // mark odd exponent of the prime number
      }
   }
  else
   {
     // this is the common case (which works also for 2)  
     unsigned int z=0;
     while (mpz_div_ui(x,y,PrimeNumbers[1])==0) { mpz_set(y,x); z++; }

     if (z&1) Relation_sparse->append(1); // mark odd exponent of the prime number
     if (z>1) // need to multiply "valid square roots" into Delta?
      {
        if (z<4) mpz_mul_ui(Delta,Delta,PrimeNumbers[1]);
        else
         {
	   mpz_set_ui(x,PrimeNumbers[1]); mpz_powm_ui(x,x,z/2,n);
	   mpz_mul(Delta,Delta,x);
         }
      }
   }

  {
    // find factors in static factorbase
    // (beginning with the 3. prime of FB (s.o) )
    vector<unsigned int> FB_Primteiler;
    GetStaticFactorHits(FB_Primteiler,SievePos);

    // multiply factors to collect them, and then divide them in one go
    mpz_set_ui(x,1);
    for (unsigned int i=0; i<FB_Primteiler.size(); i++) mpz_mul_ui(x,x,PrimeNumbers[FB_Primteiler[i]]);
    mpz_divexact(y,y,x);

    // new factor reduction using the principle of inclusion/exclusion
    // die Faktor-Potenzen nach dem Prinzip Inklusion/Exklusion abdividieren
    vector<unsigned int> WithOddExponent, TempTeiler;
    while (mpz_cmp_ui(x,1)>0) // is there a factor from our FB?
     {
       mpz_gcd(x,x,y); // prime numbers of FB, that divide a even number of times
       mpz_divexact(y,y,x); // divide by them
       mpz_mul(Delta,Delta,x); // multiply them to our square root Delta
       for (unsigned int i=0; i<FB_Primteiler.size(); ++i)
        if (mpz_divisible_ui_p(x,PrimeNumbers[FB_Primteiler[i]]))
          TempTeiler.push_back(FB_Primteiler[i]); // will divide again...
        else
          WithOddExponent.push_back(FB_Primteiler[i]); // this was the last time that it divided.
       FB_Primteiler.swap(TempTeiler); TempTeiler.clear();

       mpz_gcd(x,x,y); mpz_divexact(y,y,x); // odd number of times
       for (unsigned int i=0; i<FB_Primteiler.size(); ++i)
        if (mpz_divisible_ui_p(x,PrimeNumbers[FB_Primteiler[i]]))
          TempTeiler.push_back(FB_Primteiler[i]); // will divide again
       FB_Primteiler.swap(TempTeiler); TempTeiler.clear();
     }
    mpz_mod(Delta,Delta,n);
    std::sort(WithOddExponent.begin(),WithOddExponent.end());
    for (unsigned int i=0; i<WithOddExponent.size(); ++i)
     {
       Relation_sparse->append(WithOddExponent[i]);
     }
  }


  // reduce factors from dynamic factorbase (and maybe pushed squares...)
  // multiply the factors together and divide them in one go.
  if (HitCount)
   {
     mpz_set_ui(x,Hits[0].Faktor); Hits[0].Exponent = 1;
     for (int i=1; i<HitCount; i++) { Hits[i].Exponent = 1; mpz_mul_ui(x,x,Hits[i].Faktor); }
#if defined(DEBUG)
     if (!mpz_divisible_p(y,x)) { MARK; cerr << "Wrong divisor in Hits!" << endl; exit(1); }
#endif
     mpz_divexact(y,y,x);
     mpz_gcd(x,x,y);
     if (mpz_cmp_ui(x,1)!=0) // are there higher exponents?
      {
        // find exponents of factors in dynamic factorbase (and maybe pushed squares...)
        for (int i=0; i<HitCount; ++i)
         {
           while (mpz_divisible_ui_p(y,Hits[i].Faktor))
     	    {
     	      mpz_divexact_ui(y,y,Hits[i].Faktor);
     	      Hits[i].Exponent++; // increment exponent of factor
     	    }
#ifdef DEBUG
           if (Hits[i].Exponent==2)
            {
              MARK;
              cout << "Square in Hits[]: " << Hits[i].Faktor;
              if (is_dynamic_factor(Hits[i].Faktor))
               cout << " (dynamic factor)" << endl;
              else
               cout << " (pushed square)" << endl;
            }
#endif
           if (Hits[i].Exponent>1)
     	    {
     	      mpz_set_ui(x, Hits[i].Faktor);
     	      mpz_powm_ui(x,x,Hits[i].Exponent/2, n);
     	      mpz_mul(Delta,Delta,x); mpz_mod(Delta,Delta,n);
     	      Hits[i].Exponent%=2;
     	    }
         }
      }
   }

  // Restfactor in static factorbase...
  const bool existiert_grosser_Restfaktor = (mpz_cmp_ui(y,biggest_Prime_in_Factorbase)>0);

  if (existiert_grosser_Restfaktor && collecting_phase_finished)
    { relevant_factor=special_factor; return; } // premature cancellation

  if (!existiert_grosser_Restfaktor)
    {
      if (mpz_cmp_ui(y,1)!=0) 
        {
          cout << "CRelation::constructor:" << endl;
          cout << "Fehler in Relation / BUG im Programm:" << endl;
          cout << y << " should be 1!" << endl;
          exit(1);
        }
      relevant_factor=largest_factor_in_Relation();
      // relevant factor can be elinimated later (by utilizing this relation)
    }

#if 0 || defined(SIEVING_MORE_LARGE_SQUARES)
  // dynamic factor sanity check
  for (int i=0; i<HitCount; ++i) if(Hits[i].Exponent!=0)
    {
      if (!is_dynamic_factor(Hits[i].Faktor))
	{
	  cerr << "Inconsistency in dynamic factorbase! " << Hits[i].Faktor << " not found as known dynamic factor!" << endl;
          // This is actually an error OR the radix of a pushed square has
          // divided with an odd number of times, and the radix isn't a SLP
          // yet (so we cannot resolve the resulting relation to a static,
          // dynamic or special relation).  This shouldn't happen too often,
          // but it can happen.  Since this is unlikely, we can exit with an
          // error to watch this special case instead of just skipping the
          // relation.
	  exit(1); // not necessary, because it should also be safe just to skip this relation
          relevant_factor=dynamic_factor; return; // skip the relation
	}
    }
#else
#ifdef DEBUG
 #warning "dynamic factor sanity check disabled! (just want to mention it)"
#endif
#endif

  if (existiert_grosser_Restfaktor)
    {
      const bool is_possible_dynamic = (mpz_cmp_ui(y, SingleLargePrime_Threshold)<=0);
      if ( !is_possible_dynamic && CmpqsFactor::DLP_rejected(y) )
        {
          // We should first filter out invalid big numbers, especially the
          // primes! It doesn't harm (except for performance) if some composite
          // numbers are filtered out as primes.
          // Important is that "DLP_reject(y)" is not too expensive!
	  relevant_factor=special_factor; // ignored factor (because we do not insert any relation)
	  return;
        }
        
      if (is_possible_dynamic && probab_prime(mpz_get_ui(y)) )
        {
          // this is a new dynamic factor
          relevant_factor=dynamic_factor;
          TDynamicFactorRelation relation;
          relation.factor=mpz_get_ui(y);
          relation.SQRT_kN=SQRT_kN_mod_PrimeNumber(relation.factor);

          // the found factor could be a "non-sievable" SingleLargePrime
          // We should detect this and take appropriate action!
          if (!relation.sieveable() && is_dynamic_factor(relation.factor))
           {
             Hits[HitCount].Faktor=relation.factor; Hits[HitCount].Exponent=1; ++HitCount;
             relevant_factor=largest_factor_in_Relation();
             goto done_with_grosser_Restfaktor;
           }

#if !defined(IS_SERVER) && !defined(USE_FAKEHEAP)
          compute_Deltas_for(relation, SieveOffset+PhysicalSieveSize);
#endif
      
#ifndef IS_CLIENT
          relation.fpos = save(DynamicRelations_to_file,relation.factor, HitCount);
#else
          // send it to the output file
          relation.fpos=save(communication_stream,relation.factor, HitCount);
#endif
          DynamicFactorRelations.insert(relation);
          statistical_data::directly_sieved_DynamicFactors++;
#ifndef IS_CLIENT
          SpecialRelations::split_by_primefactor(relation.factor);
#endif
          return;
        }
        
      // it's a special factor, which isn't a prime
      {
        relevant_factor=special_factor;
        TSpecialFactorRelation relation;
        if (!relation.factor.DLP_get(y))
          {
            //cerr << "Hint: Special Factor " << y << " invalid." << endl;
            return; // Don't take over this hit, since it is invalid.
          }
#ifndef IS_CLIENT
        if (!SpecialRelations::insert(relation.factor,this,HitCount))
          {
            relevant_factor=largest_factor_in_Relation();
            // if factor hasn't been marked, then this means, that it
            // has been neutralised. -> we have found a normal relation.

            // /FIXME: nachfolgendes sollte vielleicht besser in SpecialRelations::insert() getan werden?
            //            -> in English: I'm not sure whether the following block should eventually be move into SpecialRelations::insert()
            // kombiniere Faktoren der dynamic Factorbase
            multi_combine_init();
            for (int i=0; i<HitCount; ++i) if(Hits[i].Exponent!=0)
             {
               //cout << "combining dynamic factor " << Hits[i].Faktor << endl;
               TDynamicFactorRelation FaRelSearch;
               FaRelSearch.factor=Hits[i].Faktor;
               TDynamicFactorRelations::iterator p = DynamicFactorRelations.find(FaRelSearch);
               mpz_mul_ui(Delta,Delta,(*p).factor); mpz_mod(Delta,Delta,n);
               multi_combine_main(DynamicRelations_from_file,(*p).fpos);
               //cout << "combined: " << *pos << endl;
             }
            multi_combine_exit();
            return;
          }
#else
        // send to output file
        relation.fpos=save(communication_stream, relation.factor, HitCount);
#endif
	statistical_data::directly_sieved_SpecialFactors++;
        return;
      }
    } // end of "if (existiert_grosser_Restfaktor)"
 done_with_grosser_Restfaktor:

#ifndef IS_CLIENT
  // combining factors of the dynamic factorbase
  multi_combine_init();
  for (int i=0; i<HitCount; ++i) if(Hits[i].Exponent!=0)
   {
     //cout << "combining dynamic factor " << Hits[i].Faktor << endl;
     TDynamicFactorRelation FaRelSearch;
     FaRelSearch.factor=Hits[i].Faktor;
     TDynamicFactorRelations::iterator p = DynamicFactorRelations.find(FaRelSearch);
     mpz_mul_ui(Delta,Delta,(*p).factor); mpz_mod(Delta,Delta,n);
     multi_combine_main(DynamicRelations_from_file,(*p).fpos);
     //cout << "combined: " << *pos << endl;
   }
  multi_combine_exit();
#else
  // output to file
  save(communication_stream,1, HitCount);
#endif
      
  statistical_data::DynamicFactorRating.increment_Hits_at_position(HitCount);
  // wenn DynamicFactorRating.size() auf 7 initialisiert wurde:
  // bei DynamicFactorRating[0]: Relation vollstndig durch statische FB entdeckt
  // bei DynamicFactorRating[0<k<6]: Relation durch Eliminierung von (genau) k Faktoren aus dyn. FB entdeckt
  // bei DynamicFactorRating[6]: Relation durch Eliminierung von mindestens 6 Faktoren aus dyn. FB entdeckt
  
  //cout << "relation: " << *this << endl;
  //cout << "relevant factor: " << relevant_factor << endl;
  //cout << "Finished initialization of this relation! (leaving constructor)" << endl;
}

#endif /* IS_SERVER not defined */



CmpqsFactor CRelation::combine(istream &in)
{
  //cout << "combining relation from file" << endl;
  CmpqsFactor ret;
  string s;
  int index;
 
  // save file flags
  const ios::fmtflags oldFlags = in.flags(); // memorize old flags of the stream
  in.unsetf(ios::hex); in.setf(ios::dec); // enable decimal output (notice: because of "correction factors" there could be a recursive call!)
  in.unsetf(ios::showbase); // without any leading prefix
 
  in >> s; // starting symbol "G"
  if (s != "G")
    {
      MARK;
      cerr << "error: wrong position while reading relation" << endl;
      exit(1);
    }
  in >> ret; // read factor (dynamic-factor or special-factor of relation or 1)
  in >> s; // delta
  mpz_set_str(x, s.c_str(), mpzbase_f); // x is classwide auxiliary variable
  mpz_mul(Delta, Delta, x);
  mpz_mod(Delta, Delta, n);

  // now we start to work in hexadecimal notation:
  in.unsetf(ios::dec); in.setf(ios::hex); // hexadecimal output
  in.unsetf(ios::showbase); // without "0x"-marker
  
  /*
    compute symmetric difference of the factors in the relation
    Multiply the common part (=intersection) into square root Delta
  */
  mpz_set_ui(x,1); // for speedup: collect factors in x as long as x<n
  signed int distance; // hint: we work with differences instead of the values itself

  if (Relation_dense) // Relation is dense
    { // compute symmetric difference using dense-representation
      distance=GetValue(in); index=distance;
      while (distance >= 0)
	{
	  if (Relation_dense->test(index))
	    if (index>0)
	      {
		mpz_mul_ui(x,x,PrimeNumbers[index]);
		if (mpz_cmp(x,n)>=0)
		  {
		    mpz_mod(x,x,n); mpz_mul(Delta,Delta,x);
		    mpz_mod(Delta,Delta,n); mpz_set_ui(x,1);
		  }
	      } //else mpz_neg(Delta,Delta);
	  Relation_dense->invert(index);
	  distance=GetValue(in); index+=distance;
	}
    }
  else
    { // compute symmetric difference using sparse-representation
      int p1 = 0;
      static CTinyFBsizetypeVector neueMenge(StaticFactorbase::Size());
      neueMenge.reset(); // "clear" without deallocating memory
      const CTinyFBsizetypeVector& R = *Relation_sparse;
      
      distance=GetValue(in); index=distance; 
      while (distance>=0 && p1<R.size())
	{
	  while (index<R[p1])
	    {
	      neueMenge.append(index); distance=GetValue(in); index+=distance;
	      if (distance<0) goto done;
	    }
	  while (R[p1]<index)
	    {
	      neueMenge.append(R[p1]); ++p1;
	      if (p1==R.size()) goto done;
	    }
	  while (index==R[p1])
	    {
	      if (index>0)
		{
		  mpz_mul_ui(x,x,PrimeNumbers[index]);
		  if (mpz_cmp(x,n)>=0)
		    {
		      mpz_mod(x,x,n); mpz_mul(Delta,Delta,x);
		      mpz_mod(Delta,Delta,n); mpz_set_ui(x,1);
		    }
		} //else mpz_neg(Delta,Delta);
	      distance=GetValue(in); index+=distance; ++p1;
	      if (distance<0) goto done;
	      if (p1==R.size()) goto done;      
	    }
	}
    done:
      while (distance>=0) { neueMenge.append(index); distance=GetValue(in); index+=distance; }
      while (p1!=R.size()) { neueMenge.append(R[p1]); ++p1; }
      Relation_sparse->copy_from(neueMenge); // take over new set
      //optisize(); // optize memory usage (sparse<->dense, etc.)
    }
  
  // x contains (possibly) a remaining value, which has to be multiplied into Delta:
  mpz_mul(Delta,Delta,x); mpz_mod(Delta,Delta,n);
 
  if (distance == -2) // "correction factors" follow
    {
      /* Korrekturfaktoren sind dynamische Faktoren, die zum Zeitpunkt der Generierung
	 der gelesenen Relation nicht aufgelst wurden. Sie mssen daher jetzt verknpft werden. */
      /* correction factors are dynamische factors, that weren't resolved at the time of generating
         this relation. They must be considered and combined now! */
      TDynamicFactorRelation FaRelSearch;
      stack<int> DeferredFactors; // breadth first search (Breitensuche) to avoid perpetually seeking on file.
      while ( (FaRelSearch.factor=GetValue(in))>0 ) DeferredFactors.push(FaRelSearch.factor);
       
      while (!DeferredFactors.empty())
	{
	  FaRelSearch.factor=DeferredFactors.top(); DeferredFactors.pop();
	  //cout << "Korrekturfaktor: " << FaRelSearch.factor << " fr " << setprecision(20) << ret << endl;
	  TDynamicFactorRelations::iterator p = DynamicFactorRelations.find(FaRelSearch);
	  if (p==DynamicFactorRelations.end())
	    {
	      cerr << "Unknown correction value (Korrekturfaktor) " << FaRelSearch.factor << endl;
	      exit(1); // this is the safe way to handle the error! (be partially correct!)
	      ret=0; break; // ignore and skip this, and continue with new data...
	    }
	  mpz_mul_ui(Delta,Delta,(*p).factor); mpz_mod(Delta,Delta,n);
	  if (combine(DynamicRelations_from_file,(*p).fpos).IsTypeOf(CmpqsFactor::empty)) { ret=0; break; } // auf Hintergrundspeicher kombinieren und bei Fehler ret=0 setzen
	  //cout << "finished combining " << FaRelSearch.factor << endl;
	}
    }
  
  in.flags(oldFlags); // restore memorized stream flags
 
  relevant_factor=largest_factor_in_Relation();
  //cout << "finished combining relation from file." << endl;
  return ret;
}

CmpqsFactor CRelation::multi_combine_main(istream &in)
{
  //cout << "multi-combine relation from file..." << endl;
  ++multi_combine_Counter; // count calls (do not remove; this is essential!)

  CmpqsFactor ret;
  string s;
  int index;
 
  // save file flags
  const ios::fmtflags oldFlags = in.flags(); // memorize old flags of the stream
  in.unsetf(ios::hex); in.setf(ios::dec); // enable decimal output (notice: because of "correction factors" there could be a recursive call!)
  in.unsetf(ios::showbase); // without any leading prefix
 
  in >> s; // starting symbol "G"
  if (s != "G")
    {
      MARK;
      cerr << "error: wrong position while reading relation" << endl;
      exit(1);
    }
  in >> ret; // read factor (dynamic-factor or special-factor of relation or 1)
  in >> s; // delta
  mpz_set_str(x, s.c_str(), mpzbase_f); // x is classwide auxiliary variable
  mpz_mul(Delta, Delta, x);
  mpz_mod(Delta, Delta, n);

  // now we start to work in hexadecimal notation:
  in.unsetf(ios::dec); in.setf(ios::hex); // hexadecimal output
  in.unsetf(ios::showbase); // without "0x"-marker
  
  /*
    compute symmetric difference of the factors in the relation
    Multiply the common part (=intersection) into square root Delta
  */
  signed int distance; // hint: we work with differences instead of the values itself

  if (Relation_dense) // Relation is dense
    { // compute symmetric difference using dense-representation
      distance=GetValue(in); index=distance;
      while (distance >= 0)
	{
	  if (Relation_dense->test_and_invert(index)) ++ExponentArray[index];
	  distance=GetValue(in); index+=distance;
	}
    }
  else
    { // compute symmetric difference using sparse-representation
      int p1 = 0;
      static CTinyFBsizetypeVector neueMenge(StaticFactorbase::Size());
      neueMenge.reset(); // "leeren" ohne Speicherfreigabe
      const CTinyFBsizetypeVector& R = *Relation_sparse;
      
      distance=GetValue(in); index=distance; 
      while (distance>=0 && p1<R.size())
	{
	  while (index<R[p1])
	    {
	      neueMenge.append(index); distance=GetValue(in); index+=distance;
	      if (distance<0) goto done;
	    }
	  while (R[p1]<index)
	    {
	      neueMenge.append(R[p1]); ++p1;
	      if (p1==R.size()) goto done;
	    }
	  while (index==R[p1])
	    {
	      ++ExponentArray[index];
	      distance=GetValue(in); index+=distance; ++p1;
	      if (distance<0) goto done;
	      if (p1==R.size()) goto done;      
	    }
	}
    done:
      while (distance>=0) { neueMenge.append(index); distance=GetValue(in); index+=distance; }
      while (p1!=R.size()) { neueMenge.append(R[p1]); ++p1; }
      Relation_sparse->copy_from(neueMenge); // take over new set
    }

  if (distance == -2) // "correction factors" follow
    {
      /* Korrekturfaktoren sind dynamische Faktoren, die zum Zeitpunkt der Generierung
	 der gelesenen Relation nicht aufgelst wurden. Sie mssen daher jetzt verknpft werden. */
      /* correction factors are dynamische factors, that weren't resolved at the time of generating
         this relation. They must be considered and combined now! */

      TDynamicFactorRelation FaRelSearch;
      stack<int> DeferredFactors; // breadth first search (Breitensuche) to avoid perpetually seeking on file.
      while ( (FaRelSearch.factor=GetValue(in))>0 ) DeferredFactors.push(FaRelSearch.factor);
       
      while (!DeferredFactors.empty())
	{
	  FaRelSearch.factor=DeferredFactors.top(); DeferredFactors.pop();
	  //cout << "Korrekturfaktor: " << FaRelSearch.factor << " fr " << setprecision(20) << ret << endl;
	  TDynamicFactorRelations::iterator p = DynamicFactorRelations.find(FaRelSearch);
	  if (p==DynamicFactorRelations.end())
	    {
	      cerr << "Unknown correction value (Korrekturfaktor) " << FaRelSearch.factor << endl;
              exit(1); // fr die partiell korrekte Variante...
	      ret=0; break; // Weiterarbeiten trotzdem ermglichen... 
	    }
	  mpz_mul_ui(Delta,Delta,(*p).factor); mpz_mod(Delta,Delta,n);
	  if (multi_combine_main(DynamicRelations_from_file,(*p).fpos).IsTypeOf(CmpqsFactor::empty)) { ret=0; break; } // auf Hintergrundspeicher kombinieren; im Fehlerfall ret=0 setzen
	  //cout << "finished combining " << FaRelSearch.factor << endl;
	}
    }
  
  in.flags(oldFlags); // restore memorized stream flags
 
  relevant_factor=largest_factor_in_Relation();
  //cout << "finished multi-combine from file." << endl;
  return ret;
}

streampos CRelation::save(ostream &out, const CmpqsFactor factor, const short int HitCount /* = 0 */) const
{
#ifndef IS_CLIENT
  streampos pos;
  // out.seekp(0,ios::end);  // should be already at the end of the file
  pos = out.tellp(); // get file position
#endif
  char *str;

  str = new char [mpz_sizeinbase(Delta,mpzbase_f)+2]; mpz_get_str(str,mpzbase_f, Delta);
  out << "G " << setprecision(20) << factor << " " << str << " ";
  delete [] str;

  // now we start to work in hexadecimal notation:
  const ios::fmtflags oldFlags = out.flags(); // memorize old stream flags
  out << hex; // hexadecimal output
  out.unsetf(ios::showbase); // without "0x"-marker
  
  /* remark:
     Instead of storing the values theirself, we store the distances between them (which is more space-efficient!) */
  if (Relation_dense)
    { 
      int prev = 0;
      for (int i=Relation_dense->first(); i>=0; i=Relation_dense->next(i))
	{
          PutValue(out,i-prev);
	  prev=i;
	} 
    }
  else
    {
      int prev = 0;
      for (int i=0; i<Relation_sparse->size(); ++i)
	{
 	  int h = (*Relation_sparse)[i];
          PutValue(out,h-prev);
	  prev=h;
	}
    }

  int Count_SLP_in_relation = 0;
  for (int i=0; i<HitCount; ++i) if (Hits[i].GetExponent()>0) ++Count_SLP_in_relation;
  if (Count_SLP_in_relation>0)
    {
      PutValue(out,-2);
      //cout << "Korrekturfaktoren fr Relation mit Faktor " << setprecision(20) << factor << endl;
      for (int i=0; i<HitCount; ++i)
	{ 
	  //cout << Hits[i].Faktor << "^" << Hits[i].Exponent << endl; 
	  if (Hits[i].GetExponent()>0) PutValue(out,Hits[i].Faktor);
	}
    }
  PutValue(out,-1); out << endl;
  out.flags(oldFlags); // restore old state of format-flags
#ifdef IS_CLIENT
  return 0;
#else
  return pos;
#endif
}



#ifndef IS_CLIENT
const bool CRelation::ComputeQuadraticCongruence() const
{
  /* The determined relation satisfies the congruency (A=1) mod n.
     Because of A-1=(A+1)*(A-1)=0 mod n, we can construct a factor by computing ggT(A+1,n);
     we have a good chance that this factor is not a trival one... */

  StatusReport(true);
#ifdef VERBOSE_NOTICE
  cout << endl << "Found a suitable relation..." << endl;
#endif

  if (!empty())
   {
     MARK;
     cerr << "Error! Relation contains unresolved members!" << endl;
   }

  mpz_powm_ui(x,Delta,2,n);

#ifdef VERBOSE_INFO
  cout << "[" << Delta << "] = 1 (mod N)" << endl;
  cout << "-> " << x << " = 1 (mod N)" << endl;
#endif
  
  if (mpz_cmp_ui(x,1)!=0) 
    {
      cout << "Error! Squares are NOT congruent!" << endl;
      exit(1);
    }
  
  mpz_add_ui(x,Delta,1);
  mpz_gcd(x,x,n); mpz_divexact(y,n,x); // notice: n and *not* kN is the number to factorize!
  
  if ((mpz_cmp_ui(x,1)==0) || (mpz_cmp_ui(y,1)==0))
    {
#ifdef VERBOSE_NOTICE
      cout << "factorization was trivial..." << endl << endl;
#endif
    }
  else
    {
      if (mpz_cmp(x,y)>0) mpz_swap(x,y); // we want x<y

      cout << endl << "factor found!" << endl;
      cout << x << " * ";
      cout << y << endl;
      cout << "= ";
      cout << n << endl;
      
      if (mpz_probab_prime_p(x,probab_prime_checks))
	{
	  mpz_divexact(n,n,x);
	  cout << x << " is a factor." << endl;
          std::ostringstream comment;
          comment << " [dmpqs]";
	  Factorization_to_file << MAL(x,1,comment) << flush;
	}
      else
      	if (Potenztest(x)) mpz_divexact(n,n,x); // remove powers...

      if (mpz_probab_prime_p(y,probab_prime_checks))
	{
	  mpz_divexact(n,n,y);
	  cout << y << " is a factor." << endl;
          std::ostringstream comment;
          comment << " [dmpqs]";
	  Factorization_to_file << MAL(y,1,comment) << flush;
	}
      else
	if (Potenztest(y)) mpz_divexact(n,n,y); // remove powers...
      

      if (mpz_cmp_ui(n,1)==0) // factorization complete?
	{ 
	  cout << "factorization successfully completed." << endl;
          Factorization_to_file << endl;
          return true; 
	}
    }
#ifdef VERBOSE_NOTICE
  cout << "continue factorization..." << endl << endl;
#endif
  return false;
}
#endif


// ------ pre-multiplier for MPQS -----------
#include "Vorfaktorbestimmung.cc"
// ------------------------------------------


#ifndef IS_CLIENT
 #include "parse_term.cc"
 // this is a predicate stating "false"
 const bool without_multi_combine_init = false; // true ->multi-init, false ->no multi_init on StaticRelations::insert
#endif

#include "SpecialRelations.cc"


#ifndef IS_SERVER
 // this part cannot be compiled separately and must therefore be included
 #include "Sieving-inc.cc"
#endif


#ifdef IS_STANDALONE
void process_ecm()
{
  streampos fpos = Recovery_from_file.tellg();
  int Anzahl_Kurven = 0;
  string s;

  srand(time(NULL)); // setting a random seed (time(NULL) should satisfy for our needs)
  
  unsigned int dezimalstellen = mpz_sizeinbase(n,10);
  tune_parameters(dezimalstellen); // as the name says: tuning of parameters according to our input number

  Recovery_from_file >> Anzahl_Kurven;
  Recovery_from_file.ignore(1,'\n'); // overread "end of line"

  {
   ifstream in("ecm-recover.dat");
   if (in) // does a recovery file exist?
    {
     in.close();
     elliptic_curves elliptic_curve; // create a curve
     elliptic_curve.go(-1,elcu_Phase1,static_cast<elliptic_curves::NATNUM>(elcu_Phase2)); // try to recover curve and retry factorization with elliptic curve
    }
  }

  while (Anzahl_Kurven<elcu_Kurven)
  {

    Anzahl_Kurven++;
    cout << "elliptic curve #" << Anzahl_Kurven << "/" << elcu_Kurven << endl;

    int ecm_sigma = rand(); // choose a random curve
    elliptic_curves elliptic_curve; // create curve
    elliptic_curve.go(ecm_sigma,elcu_Phase1,static_cast<elliptic_curves::NATNUM>(elcu_Phase2)); // try factorization with elliptic curve

    // write number of processed curves to recovery file
    Recovery_to_file.seekp(fpos); 
    Recovery_to_file << Anzahl_Kurven << endl << flush;

    // found a factor?
    if (dezimalstellen==mpz_sizeinbase(n,10)) continue; // no new factor
    
    // digits of n has changed -> new factor

    if (mpz_probab_prime_p(n,probab_prime_checks))
      {
        Factorization_to_file << MAL(n);
        cout << "remaining factor: " << n << endl;
        cout << "remaining factor is most probably prime!" << endl;
        mpz_set_ui(n,1); // n has been factorized
      }

    if ( (mpz_cmp_ui(n,1)==0) || Potenztest(n) ) // factorization complete?
      { 
        cout << "factorization successfully completed." << endl;
        Factorization_to_file << endl;
        ExitManager::StopFactorization(); // terminate program (do finalize work)
      }

    // tune parameters for further ecm-rounds...
    dezimalstellen=mpz_sizeinbase(n,10);
    tune_parameters(dezimalstellen);

    // since the number to factorize has become smaller, the recovery-file need to be updated as well!
    Recovery_buffer.close(); // close file
    Recovery_buffer.open(RecoveryFile.c_str(),ios::in|ios::out|ios::trunc); // re-open (-> create empty file)
    Recovery_to_file << n << endl; // write out number to factorize
    fpos = Recovery_from_file.tellg(); // mark position
    Recovery_to_file << Anzahl_Kurven << endl << flush; // store number of curves done so far
    ecm_curves_processed=Anzahl_Kurven;

  }
  cout << "ECM-Phase (standalone) completed." << endl;
}
#endif



#ifdef IS_SERVER
 #include "Cprocess_clients.cc"
 #include "XML_StatusServer.cc"
#endif /* IS_SERVER */


#if defined(IS_CLIENT) && defined(USE_NETWORK)
 #include "Client_IO.cc"
#endif


#ifdef USE_NETWORK
void kill_all_other_threads()
{
  //cout << "Killing explicitely all other threads..." << endl;
  //pthread_kill_other_threads_np();
}
#endif

void cleanup_files()
{
#ifdef VERBOSE_INFO
  cout << "cleaning files..." << endl;
#endif
  StaticRelations::cleanup_files();
  SpecialRelations::cleanup_files();

  // close streams
  DynamicRelations_buffer.close();
  Recovery_buffer.close();

  // delete files
  remove(DynamicRelationsFile.c_str());
  remove(RecoveryFile.c_str());
}

void cleanup_memory()
{
#ifdef VERBOSE_INFO
  cout << "cleanup allocated memory" << endl;
#endif
  StaticRelations::cleanup_memory();
#ifdef VERBOSE_INFO
  cout << "cleanup mpz numbers" << endl;
#endif
  mpz_clear(kN); mpz_clear(n);
#ifdef VERBOSE_INFO
  cout << "mpz numbers cleared." << endl;
#endif
}


#ifdef IS_SERVER
 void signalhandler(int signr)
 {
#ifdef VERBOSE_WARN
   cout << "Signal " << signr << " received. (ThreadId=" << pthread_self() << ")" << endl;
#endif
   exit(1);
 }
#endif


#ifdef USE_NCURSES
class Cncursed : private ForbidAssignment
{
private:
  static Cncursed* myself;
  static void suicide() { delete myself; }

  typedef streambuf* Pstreambuf;
  typedef Cwin* Pwin;
  Pstreambuf cout_titel_buf, cout_network_buf, cout_status_buf,
             cerr_buf, clog_buf, cout_buf;
  Pwin window1, window2, window3;
public:
 Cncursed()
 {
   if (myself)
    {
      MARK;
      cerr << "You may not call me twice, I'm unique!" << endl;
      exit(93);
    }
   myself=this; // I am myself, really!
   init_my_ncurses();
   window1 = new Cwin(8,0,0,0);
   window2 = new Cwin(8,0,8,0);
   window3 = new Cwin(0,0,16,0);
   cout_titel_buf=cout_titel.rdbuf(window1->rdbuf());
   cout_network_buf=cout_network.rdbuf(window1->rdbuf());
   cout_status_buf=cout_status.rdbuf(window2->rdbuf());
   cerr_buf=cerr.rdbuf(window2->rdbuf());
   clog_buf=clog.rdbuf(window2->rdbuf());
   cout_buf=cout.rdbuf(window3->rdbuf());
   atexit(suicide);
 }
 ~Cncursed()
 {
   cout_titel.rdbuf(cout_titel_buf);
   cout_network.rdbuf(cout_network_buf);
   cout_status.rdbuf(cout_status_buf);
   cerr.rdbuf(cerr_buf);
   clog.rdbuf(clog_buf);
   cout.rdbuf(cout_buf);
   delete window1; delete window2; delete window3;

   exit_my_ncurses();

#if 0
   // print statistical information before exiting,
   // since screen should be cleared now!
   display_StatusLegend();
   StatusReport(true);
#endif
 }
};
Cncursed* Cncursed::myself = NULL;
#endif


int main(const int argc, const char* const argv[])
{

#ifdef USE_NETWORK
 atexit(kill_all_other_threads);
#endif

#ifdef USE_NCURSES
  new Cncursed(); // trigger activation of ncursed streams
#endif

  cout.setf(ios::fixed); // Dezimaldarstellung, nicht Scientific!

 {
   string name = "Faktorisierungsprogramm ";
   name+=VERSION;
   string author = "written 2004-11 by Thorsten Reinecke";
   string email = "(thre@thorstenreinecke.de)";
   const int L = 5+MAX(name.length(),MAX(author.length(),email.length()));
   const string sep(L,'-');
   int f;
   f=(L-name.length())>>1; name=string(f,' ')+name+string(f+1,' '); name.resize(L);
   f=(L-author.length())>>1; author=string(f,' ')+author+string(f+1,' '); author.resize(L);
   f=(L-email.length())>>1; email=string(f,' ')+email+string(f+1,' '); email.resize(L);
   cout_titel << " -" << sep << "-" << endl;
   cout_titel << "- " << name << " -" << endl;
   cout_titel << "- " << author << " -" << endl;
   cout_titel << "- " << email << " -" << endl;
   cout_titel << " -" << sep << "-" << endl;
 }
#ifdef __GNUG__
  {
    const string s=__VERSION__;
    cout << "(" << __DATE__ << ", " << __TIME__ << " using ";
    if (s[0]<='9') cout << "gcc ";
    cout << __VERSION__ << ")" << endl;
  }
#endif
  cout << endl;
  cout << "This program may be freely distributed under the" << endl;
  cout << "terms of the GNU-Public-Licence." << endl;
  cout << endl;
    
  if (argc>2)
    {
#ifdef IS_CLIENT
       cerr << "servername expected!" << endl;
#else
      cerr << "number for factorization expected!" << endl;
#endif
      exit(1);
    }
  
  const bool recover = (argc==1); // without argument: Recovery-Mode

  {
   // resolve ConfigFile
   if (getenv("QSIEVE_CFG"))
    {
      ConfigFile=getenv("QSIEVE_CFG");
#ifdef VERBOSE_NOTICE
      cout << "Setting ConfigFile to \"" << ConfigFile << "\"" << endl;
#endif
    }
   // and convert it to real/absolute path
#ifdef PATH_MAX
   char resolved_path[PATH_MAX];
#else
   char resolved_path[4096];
#endif /* PATH_MAX */
   if (realpath(ConfigFile.c_str(),resolved_path))
    ConfigFile=resolved_path;
   else
    {
      cerr << "Cannot find ConfigFile " << "\"" << ConfigFile << "\"" << endl;
      // server & standalone versions need a configuration file,
      // clients do not need it (but they make use of it as well)
#ifndef IS_CLIENT
      exit(1);
#endif
    }
  }

  mpz_init(n); // our number to factorize
  mpz_init(kN);
  ExitManager::register_exithandler(cleanup_memory); // on sucessful exit free allocated data

  Read_ConfigFile();

#ifndef IS_CLIENT

#ifdef IS_SERVER
  // establish the network service at this point,
  // to allow clients to connect
  const connection_waiter my_connection_waiter(server_port);
#endif

  if (!recover) // not in recovery-mode -> evaluate arguments
    { 
      // get and evaluate given number
      char* const neuer_str = strdup(argv[1]);
      char* str = neuer_str; 
      if (!parse_term::get_number(n, str))
	{
	  cout << "Wrong input at: '" << str << "'" << endl;
	  exit(1);
	}
      else
	if (str[0]!='\0')
	  {    
	    cout << "Syntax error in input term. (parenthesis?)" << endl;
	    exit(1);
	  }
	else
	  if (mpz_cmp_ui(n,0)<=0)
	    {
	      cout << "Input must be positive natural number!" << endl;
	      exit(1);
	    }
      free(neuer_str); // don't call "delete []" because of "stdup/malloc"
      FoundFactors.regarding=argv[1];
    }

  if (recover)
    {
      // Recovery: continue a factorization, which was previously aborted
      // (try to) open the necessary file streams
      StaticRelations::FileBuffer.open(StaticRelationsFile.c_str(),ios::in|ios::out);
      SpecialRelations::FileBuffer.open(SpecialRelationsFile.c_str(),ios::in|ios::out);
      if (!DynamicRelations_buffer.open(DynamicRelationsFile.c_str(),ios::in|ios::out))
       {
         cerr << "Cannot access " << DynamicRelationsFile << endl;
         exit(1);
       }
      Recovery_buffer.open(RecoveryFile.c_str(),ios::in|ios::out|ios::ate);

      // offenbar wird trotz ios::ate nicht (immer) ans Ende positioniert
      // deshalb wird die Testabfrage modifiziert:
      if (Recovery_from_file) Recovery_from_file.seekg(0,ios::end);
#ifdef STL_STREAM_workaround
      if ( (!Recovery_from_file) || (Recovery_from_file.tellg()==std::streampos(0)) || (Recovery_from_file.tellg()==std::streampos(-1)) )
// tellg()==0 indicates empty file -> we cannot recover
// tellg()==-1 indicates failure -> we cannot recover
// remark:
//  in gcc 3.4 (cvs-experimental-2003-10-17 we cannot compare with "<" !!)
//  do we really need a workaround to check this condition? 
#else
      if ( (!Recovery_from_file) || (Recovery_from_file.tellg()<1) )
#endif /* STL_STREAM_workaround */
	{
	  cerr << "Recovery not possible!" << endl;
	  exit(1);
	}
      
      Recovery_from_file.seekg(0,ios::beg);
      Recovery_to_file.seekp(0,ios::beg);
      Recovery_from_file >> n; // Zahl holen
      Recovery_from_file.ignore(1,'\n');

      cout << "Continue factorizaction for " << endl;
      cout << n << endl;

      // mark recovery-mode in Factorization-File! One never knows, what has
      // happened to that file in the mean time!
      Factorization_to_file << " [RECOVERY] "; // flush not necessary... 
      FoundFactors.AutoLoad();
    }
  else
    {
      // First start of this factorization (i.e. not a recovery)
      // open streams
      StaticRelations::FileBuffer.open(StaticRelationsFile.c_str(),ios::out|ios::trunc);
      SpecialRelations::FileBuffer.open(SpecialRelationsFile.c_str(),ios::in|ios::out|ios::trunc);
      if (!DynamicRelations_buffer.open(DynamicRelationsFile.c_str(),ios::in|ios::out|ios::trunc))
       {
         cerr << "Cannot access " << DynamicRelationsFile << endl;
         exit(1);
       }

      Recovery_buffer.open(RecoveryFile.c_str(),ios::in|ios::out|ios::trunc);
      
      cout_status << "Starting factorization for" << endl;
      cout_status << n << endl;
      
      Factorization_to_file << endl << "Factorization of " << argv[1] << ":" << endl;
      Factorization_to_file << n << " = " << endl;
      
      easy_factor(); // remove "easy" detectable factors
      
#ifdef VERBOSE_NOTICE
      cout_status << "Starting factorization with ECM and MPQS for" << endl;
      cout_status << n << endl;
#endif
      Recovery_to_file << n << endl;
      unlink("ecm-recover.dat"); // remove any existent ecm recovery file
    }
  ExitManager::register_exithandler(cleanup_files); // on sucessful exit delete the files
#ifdef IS_SERVER
  signal(SIGINT, signalhandler);
  tune_parameters(mpz_sizeinbase(n,10));

#ifdef NOTIFY_PARENT
  // notify parent that now the distributed phase starts
  // and the server will process client connections
  kill(getppid(),SIGUSR1); // send SIGUSR1 to the parent process
#endif

  CXML_StatusServer::install_XML_StatusServer();
  if (!SkipECM)
   {
#ifdef VERBOSE_NOTICE
     cout << "Waiting for net-clients (ECM)..." << endl;
#endif
     Cprocess_clients::process_data_stream_ecm(my_connection_waiter); // start distributed ECM...
   }
#endif /* IS_SERVER */

#ifdef IS_STANDALONE
  // The standalone-version should use elliptic curves, too.
  if (!SkipECM) process_ecm(); // start ECM...
#endif /* IS_STANDALONE */

#if defined(USE_DFT) || (CONTINUATION_METHOD==2)
  // at this place, all possible functions that made possible use of
  // discrete fast fourier transform have been finished.
  //  --> we can release the DFT-object and assoziated auxiliary memory...
  polynomial::clear_dft_tempmemory();
#endif /* USE_DFT */

  // now prepare everything for the Quadratic Sieve...

  determine_best_kN(); // calculate a good/optimal value for fastest possible sieving

  /* (to be discussed)
     Should we tune the sieve parameters according to N or to kN?

     Streitfall:
     Soll man nach Ermittlung des bestem Multipliers die Siebparameter
     noch tunen?
     Ja   -> denn die Zahl kN hat ja die entsprechende Gre...
     Nein -> denn fr die Wahl des Multipliers wurden diese Gren nicht
             bercksichtigt und entscheidend sollte die Gre der
             zu faktorisierenden Zahl sein. Nur so ist eine Vergleichbarkeit
             der Multiplier unter gleichen Voraussetzungen zu gewhrleisten.
    Wie auch immer:
     Eine "optimale" Ermittlung der Siebgre und der Faktorbasis kann nur
     mittelbar von der Gre der Zahl abhngen. Die Laufzeit der Siebphase
     hngt davon ab, wie viele Relationen zu ermitteln sind *und* wie
     schnell die Relationen ermittelt (=gesiebt) werden knnen!
     Letzterer Wert ist funktional von der Gre der Zahl und der Qualitt
     der Faktorbasis abhngig. Diese Qualitt ist zwar durch Vorgabe eines
     Multipliers (nach Annahme) maximiert worden; aber das heit noch nicht,
     dass dadurch die funktionale Abhngigkeit beseitigt wre.
     Mit anderen Worten: Wenn wir die Bestimmung der Faktorbasisgre und
     der Siebgre *nur* von der Gre der Zahl abhngig machen, so gehen
     wir idealisierend von einer einheitlichen Qualitt der Faktorbasis aus.
     Da wir diese einheitliche Qualitt der Faktorbasis nicht definiert haben,
     besitzen wir auch keine Entscheidungsgrundlage darber, ob fr unsere
     Zahl (n bzw. kN) eine unterdurchschnittliche, durchschnittliche oder
     berdurchschnittliche Faktorbasisqualitt besteht. Aus dieser Sichtweise
     erscheint der obige Streitfall nicht entscheidbar zu sein.
  */

#if 0   
  tune_parameters(mpz_sizeinbase(kN,10)); // tune parameters for kN
#else
  tune_parameters(mpz_sizeinbase(n,10)); // tune parameters for n
#endif
  /* Warning:
      Be consistent in the distributed version!
      A client should get its parameters from the server.
      Different cfg-files on different platforms would otherwise lead to problems.
  */

#else /* IS_CLIENT */
  if (recover)
    {
      cerr << "Missing parameter" << endl;
      exit(1);
    }
  communication_name = argv[1];
#ifdef USE_NETWORK
  CClientRelation_Delivery::init();

  { // first execute the elliptic curves (net-client):

    if (recover)
     {
      // in recovery mode we should try to continue an aborted ECM, if possible 
      ifstream in("ecm-recover.dat");
      if (in) // does a recovery file exist?
       {
        in.close();
        elliptic_curves elliptic_curve; // create curve
        elliptic_curve.go(-1,elcu_Phase1,static_cast<elliptic_curves::NATNUM>(elcu_Phase2)); // try to recover curve and retry factorization with elliptic curve
       }
     }

    int ecm_sigma = -1, runde=0;
    while (true)
     {
       unix_io_stream communication_stream(communication_name, server_port);
 waitloop:
       communication_stream << "ECM?" << endl;
       communication_stream >> ecm_sigma;
       if (ecm_sigma==-3)
        {
          // -3 is special token for idle waiting loop
          string s;
          communication_stream >> s;
          cout << s << endl;
          goto waitloop;
        }

       if (ecm_sigma<0) break; // sigma<0 -> curves are done!
       communication_stream >> elcu_Phase1 >> elcu_Phase2;
       communication_stream >> n; // get number
       cout << "elliptic curve " << ++runde << endl;
       elliptic_curves elliptic_curve; // create ecm object
       elliptic_curve.go(ecm_sigma,elcu_Phase1,static_cast<elliptic_curves::NATNUM>(elcu_Phase2)); // try factorization using this curve
     }
  }
  
  {
    unix_io_stream communication_stream(communication_name, server_port);
    communication_stream << "Hunger!" << endl;
    communication_stream >> n;
    communication_stream >> StaticRelations::Size_StaticFactorbase;
    communication_stream >> Factor_Threshold;
    communication_stream >> LogicalSieveSize;
#ifdef VERBOSE_NOTICE
    cout << "n=" << n << endl;
    cout << "Size of static Faktorbase: " << StaticFactorbase::Size() << endl;
    cout << "Factor-Threshold (Exponent): " << Factor_Threshold << endl;
    cout << "Sieveinterval per polynomial: [" << -LogicalSieveSize << "," << LogicalSieveSize << "]" << endl;
#endif
  }
  {
    string name;
    name = CClientRelation_Delivery::file_name+"A";
    communication_stream.open(name.c_str(), ios::out|ios::trunc);
    if (!communication_stream)
     {
       MARK; cerr << "opening communication_stream \"" << name << "\" failed!" << endl;
       exit(1);
     }
  }
  atexit(CClientRelation_Delivery::cleanup_files); // Attention: If program crashes, then the relations which are not transmitted, are gone forever!
#else /* no USE_NETWORK */
  // does the file exist?
  ifstream communication_stream2(communication_name.c_str());
  if (!communication_stream2)
    {
      cerr << "Unable to open " << communication_name << endl;
      exit (1);
    }
  communication_name+=".out";
  
  communication_stream2 >> n; 
  communication_stream2 >> StaticFactorbase::Size_StaticFactorbase;
  communication_stream2 >> Factor_Threshold;
  communication_stream2 >> LogicalSieveSize;
  communication_stream2.ignore(1,'\n');
#ifdef VERBOSE_NOTICE
  cout << "n=" << n << endl;
  cout << "Size of static Faktorbase: " << StaticFactorbase::Size() << endl;
  cout << "Factor-Threshold (Exponent): " << Factor_Threshold << endl;
  cout << "Sieveinterval per polynomial: [" << -LogicalSieveSize << "," << LogicalSieveSize << "]" << endl;
#endif
  communication_stream.open(communication_name.c_str(), ios::out|ios::trunc);
  if (!communication_stream)
   {
     MARK; cerr << "opening communication_stream \"" << communication_name << "\" failed!" << endl;
     exit(1);
   }
#endif /* USE_NETWORK */
  determine_best_kN(); // optimal multiplier value to sieve with
#endif /* not IS_CLIENT */
  
  if ( sqrt(mpz_get_d(kN)) < PhysicalSieveSize )
    {
      cerr << "Sieve size too big (you may want to reduce its size)!" << endl;
      exit(1);
    }

  // Set a threshold for Double-Large-Primes,
  // this is the square of the maximal Single Large Prime...
  mpz_init(CmpqsFactor::DLP_Threshold);
  mpz_set_ui(CmpqsFactor::DLP_Threshold,SingleLargePrime_Threshold);
  mpz_mul(CmpqsFactor::DLP_Threshold,CmpqsFactor::DLP_Threshold,CmpqsFactor::DLP_Threshold);
  
  compute_StaticFactorbase();
#ifdef VERBOSE_INFO
  cout << "The first 20 members of the static factorbase are" << endl;
  for (int i=0; i<20; i++) cout << PrimeNumbers[i] << " "; cout << endl;
  cout << "Biggest prime in static factorbase: " << biggest_Prime_in_Factorbase << endl;
  cout << "#squares (of primes) in FB: " << FB_maxQuadrate << endl;
  cout << "#powers (of primes) in FB: " << NumberOf_more_PrimePowers << endl;
#endif

  // the server does not sieve anything, so initialize the sieve only, if
  // the program is not compiled in server mode
#ifndef IS_SERVER
  SieveControl::compute_SieveThreshold(); // Threshold for detecting relations during sieving phase
  for (int i=0; i<64; ++i) SieveArray_[i]=-1; // initialize array underflow trigger values
#endif


  Polynom.compute_first_polynomial(kN,LogicalSieveSize); // compute the first MPQS polynomial to sieve with

#if defined(IS_CLIENT) && !defined(USE_NETWORK)
 // file-client: write kN to file
 communication_stream << kN << endl;
#endif
  
#ifndef IS_CLIENT
  if (recover)
    {
      // read in Static Relations
#ifdef VERBOSE_NOTICE
      cout << "Reading static relations from file..." << endl;
#endif
      StaticRelations::Load();
#ifdef VERBOSE_NOTICE
      cout << "Recover: " << StaticRelations::Count() << " static relations read." << endl;
#endif

      // read in Dynamic Relations
#ifdef VERBOSE_NOTICE
      cout << "Reading dynamic relations from file..." << endl;
#endif
      DynamicRelations_from_file.seekg(0,ios::beg);
      while (DynamicRelations_from_file.peek()=='G') 
	{
	  TDynamicFactorRelation relation;
	  relation.fpos = DynamicRelations_from_file.tellg();
	  string s; DynamicRelations_from_file >> s;
	  DynamicRelations_from_file >> relation.factor;
	  relation.SQRT_kN=SQRT_kN_mod_PrimeNumber(relation.factor);
	  DynamicRelations_from_file.ignore(1000000,'\n');
	  DynamicFactorRelations.insert(relation);
	}
      DynamicRelations_from_file.clear();
      DynamicRelations_to_file.seekp(0, ios::end); // as a precaution...
#ifdef VERBOSE_NOTICE
      cout << "Recover: " << DynamicFactorRelations.size()
	   << " dynamic relations read." << endl;
#endif

      SpecialRelations::Load();
      SpecialRelations::CycleSearch();
      
      
#ifdef IS_SERVER
      // beim verteilten Sieben knnten den Clients einige potentielle dynamische Faktoren
      // durch die Lappen gegangen sein. Dies geschieht, falls ein Client einen Special-Factor
      // findet, der nicht mit dem vollen Satz der aktuellen (und daher nur dem Server bekannten)
      // dynamischen Faktoren gesiebt wurde.
      // Anstelle eines Tests zur Laufzeit werden diese im Normalfall eher seltenen Flle
      // nun hier beim Neu-Hochfahren des Servers bercksichtigt:
      // WICHTIG: erst mssen alle dynamischen Faktoren eingelesen sein! Ansonsten knnen
      //          doppelt auftretende dynamische Faktoren nicht zu einer Relation reduziert werden!
      // folgende Anweisungen wrden das bewerkstelligen, aber es lohnt sich kaum und dauert (zu) lange:
#if 1 /* 0: wir lassen es, 1: wir machen es (trotzdem) */
      cout << "Trying to discover previously hidden SLP in special factors." << endl;
      for (TDynamicFactorRelations::iterator pos = DynamicFactorRelations.begin();
	   pos!=DynamicFactorRelations.end(); ++pos)
	SpecialRelations::split_by_primefactor((*pos).factor);
      cout << "discovered dynamic factors: " << statistical_data::Special_to_dynamic_Factor_hit << endl;
#endif
#endif
      
    }
#else // is client:
#ifndef USE_NETWORK 
  TDynamicFactorRelation relation;
  relation.fpos = 0; relation.factor=0;
  int diff; // we read the differences between factors instead the factors itself
  communication_stream2 >> diff; // read over filepos
  while (true)
    {
      communication_stream2 >> diff;
      relation.factor+=diff;
      if (diff<=0) break;
      relation.SQRT_kN=SQRT_kN_mod_PrimeNumber(relation.factor);
      DynamicFactorRelations.insert(relation);
    }
#endif
#endif  
  
  
  
#ifndef IS_CLIENT
  if (recover)
    {
      streampos fpos = Recovery_from_file.tellg();
      Polynom.load_if_available(Recovery_from_file); // get current polynomial (to continue factorization)
      Recovery_to_file.seekp(fpos);
    }
#endif

display_StatusLegend();

#ifdef IS_SERVER
#ifdef VERBOSE_NOTICE
  cout << "Waiting for clients (MPQS)..." << endl;
#endif
  //cout << "I'm the main thread with id = " << pthread_self() << endl;
#if 0
  signal(SIGINT,  signalhandler);
  signal(SIGTSTP, signalhandler);
  signal(SIGQUIT, signalhandler);
  signal(SIGTTIN, signalhandler);
  signal(SIGTTOU, signalhandler);
  signal(SIGHUP,  signalhandler);
  signal(SIGCHLD, signalhandler);
#endif
  while (true)
   {    
     int* pdata = new int;
     *pdata = my_connection_waiter.get_new_client_socket_descriptor();
     pthread_attr_t detached_thread;
     pthread_t thread_process_client_data;

     pthread_attr_init(&detached_thread);
     pthread_attr_setdetachstate(&detached_thread, PTHREAD_CREATE_DETACHED);

     //cout << "creating thread..." << endl;
     const int retcode = pthread_create(&thread_process_client_data,&detached_thread,
                          Cprocess_clients::THREAD_process_data_stream, pdata);
     if (retcode != 0)
      {
        cerr << "pthread_create failed!" << endl;
	delete pdata;
        exit(1);
      }
      
     // collecting_phase_finished= ( StaticRelations::Count() > 0.6*Size_StaticFactorbase );
     /*
        The flag "collecting_phase_finished" can be uesed to suppress the
        collection of dynamic and special factors at any time. One can do
        that if barely enough main memory or secondary memory is available
        to continue the factorization. Also it can be possible that the
        collection of these factors has proceeded so far, that sieving slows
        actually down...

        But: You cannot determine the current state of available memory for
        a UNIX machine. (Yes, there may be tricks to do that, but it is
        senseless!) At any time any process can try to free/allocate memory! 
        You simply can't control this (except as root).

        There are better ways to finetune the behaviour of Qsieve than to
        set "collecting_phase_finished"!  For example, you can abort and
        restart the program, thereby triggering a recovery, where memory
        consumption also gets optimized. You can even shorten the files
        before restarting the program.
      */
   }

// Server
#else
// not Server


#ifdef IS_CLIENT
  mpz_t UpperBound_D;
  mpz_init(UpperBound_D);
    
#ifdef USE_NETWORK

 ClientMainLoop: // main loop: get MPQS interval and sieve them

  CClientPolynomFetcher::fetch(UpperBound_D);

#else /* no network */
  Polynom.load(communication_stream2);
  communication_stream2 >> UpperBound_D;
  if (communication_stream.fail())
   {
     cerr << "No more data from file??" << endl;
     exit(1);
   }
#endif /* USE_NETWORK else no network */
#endif /* Client */
  
  // The ultimate loop:

  do // sieve relations as long no factorization has been found.
    {
#ifdef USE_NETWORK
      if (StaticRelations::Count() > 4*StaticFactorbase::Size())
       {
          cerr << "#static relations > 4*StaticFactorbase::Size(): this is weird." << endl;
          cerr << "I will sleep 30 seconds for transmission to end, and then I will abort." << endl;
          for (int i=0; i<15; ++i)
           {
             CClientRelation_Delivery::go();
             sleep(2);
           }
          exit(1);
       }
      CClientDynamicFactorFetcher::fetch();
#endif

#ifndef IS_CLIENT
      // for Recovery:
      streampos fpos = Recovery_to_file.tellp();
      Polynom.save(Recovery_to_file); // store current MPQS polynomial (for recvovery)
      Recovery_to_file.seekp(fpos);
#endif

      SieveOffset=-LogicalSieveSize;
      recompute_all_Deltas();
#ifdef USE_FAKEHEAP
      Sieve_Delta_Heap.sort();
#endif
      CSieveStaticFactors::init_LocalDeltas();
      while (SieveOffset<LogicalSieveSize)
	{
	  initialize_Sieve();
	  CSieveStaticFactors::do_sieving_StaticFactors(); // with primes out of the static factorbase
	  do_sieving_Squares(); // (bigger) squares of primes of the static factorbase
	  do_sieving_DynamicFactors(); // even more quadratic residues outside the factorbase (dynamic factors)
	  do_scanning_Sieve();
	  SieveOffset+=PhysicalSieveSize;
	}
      StatusReport(); // issue information to screen (don't remove this, it triggers some other stats, too!)

#if defined(IS_CLIENT) && defined(USE_NETWORK)
      CClientRelation_Delivery::go();
#endif  // Client && Network

      Polynom.compute_next_polynomial();
      
     // collecting_phase_finished= ( StaticRelations::Count() > 0.6*Size_StaticFactorbase );
     /*
        The flag "collecting_phase_finished" can be uesed to suppress the
        collection of dynamic and special factors at any time. One can do
        that if barely enough main memory or secondary memory is available
        to continue the factorization. Also it can be possible that the
        collection of these factors has proceeded so far, that sieving slows
        actually down...

        But: You cannot determine the current state of available memory for
        a UNIX machine. (Yes, there may be tricks to do that, but it is
        senseless!) At any time any process can try to free/allocate memory! 
        You simply can't control this (except as root).

        There are better ways to finetune the behaviour of Qsieve than to
        set "collecting_phase_finished"!  For example, you can abort and
        restart the program, thereby triggering a recovery, where memory
        consumption also gets optimized. You can even shorten the files
        before restarting the program.
      */
    }
#ifndef IS_CLIENT
  while (true);
#else
  while (Polynom < UpperBound_D);
#ifdef USE_NETWORK
  goto ClientMainLoop; // main loop: get new MPQS interval and continue sieving
#endif
  mpz_clear (UpperBound_D);
#endif
  
#endif /* not server */

#ifdef VERBOSE_INFO  
  cout << endl << "Session ended successfully." << endl;
#endif
  return 0;
}
